"""Parquet row-group indexing utilities for row streaming."""

from __future__ import annotations

from bisect import bisect_right
from dataclasses import dataclass
from threading import RLock

import polars as pl


@dataclass(slots=True, frozen=True)
class ParquetRowGroupIndex:
    path: str
    row_group_rows: tuple[int, ...]
    row_group_offsets: tuple[int, ...]
    total_rows: int

    @classmethod
    def build(cls, path: str) -> ParquetRowGroupIndex | None:
        try:
            metadata = pl.read_parquet_metadata(path)
        except Exception:
            return None

        groups = getattr(metadata, "row_groups", None)
        if not groups:
            return None

        rows: list[int] = []
        for group in groups:
            count = getattr(group, "num_rows", None)
            if count is None:
                count = getattr(group, "rows", None)
            if count is None:
                continue
            try:
                rows.append(max(0, int(count)))
            except (TypeError, ValueError):
                continue

        if not rows:
            return None

        offsets: list[int] = []
        running = 0
        for count in rows:
            offsets.append(running)
            running += count

        total_rows = getattr(metadata, "num_rows", None)
        if total_rows is None:
            total_rows = running
        try:
            total_rows = int(total_rows)
        except (TypeError, ValueError):
            total_rows = running

        return cls(path, tuple(rows), tuple(offsets), max(0, total_rows))

    def seek(self, row: int) -> tuple[int, int]:
        if row <= 0:
            return 0, 0
        index = bisect_right(self.row_group_offsets, row) - 1
        index = max(0, min(index, len(self.row_group_offsets) - 1))
        return index, self.row_group_offsets[index]


class ParquetRowGroupIndexer:
    def __init__(self, *, lock: RLock) -> None:
        self._lock = lock
        self._indices: dict[str, ParquetRowGroupIndex] = {}

    def get(self, path: str) -> ParquetRowGroupIndex | None:
        with self._lock:
            cached = self._indices.get(path)
        if cached is not None:
            return cached

        index = ParquetRowGroupIndex.build(path)
        if index is None:
            return None

        with self._lock:
            self._indices[path] = index
        return index

    def clear(self) -> None:
        with self._lock:
            self._indices.clear()


__all__ = ["ParquetRowGroupIndex", "ParquetRowGroupIndexer"]



"""Search and modal flows for the TUI screen."""

from __future__ import annotations

import os
import subprocess
from contextlib import suppress
from pathlib import Path
from typing import Any

from prompt_toolkit.application import run_in_terminal
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.widgets import Dialog, TextArea

from ..core.sheet_traits import resolve_sheet_traits
from ..core.viewer.ui_hooks import NullViewerUIHooks
from ..data.filter_lang import FilterError
from ..data.transform_lang import TransformError
from ..sheets.file_browser_sheet import file_browser_status_text
from . import modals as tui_modals
from .completions import ColumnNameCompleter, FilesystemPathCompleter


class ScreenSearchController:
    """Owns search/modals and associated histories."""

    def __init__(
        self,
        *,
        screen: Any,
        history_max_size: int,
        format_expr_filters: Any,
    ) -> None:
        self._screen = screen
        self._history_max_size = history_max_size
        self._format_expr_filters = format_expr_filters
        self._search_history: list[str] = []
        self._row_search_history: list[str] = []
        self._filter_contains_history: list[str] = []
        self._expr_filter_history: list[str] = []
        self._sql_filter_history: list[str] = []
        self._transform_history: list[str] = []
        self._command_history: list[str] = ["write output.parquet"]
        self._shell_history: list[str] = []
        self._col_search_history: list[str] = []

    def open_filter_modal(self, event, *, initial_text: str | None = None) -> None:
        screen = self._screen
        title = "Expression Filter"
        prompt_text = "Polars expression (use c.<column>) - Enter: replace existing"
        current_expr_filter = ""
        with suppress(Exception):
            current_expr_filter = self._format_expr_filters(getattr(screen.viewer, "filters", ()))
        default_text = initial_text or current_expr_filter or ""
        if not default_text:
            current_col = None
            with suppress(Exception):
                current_col = screen.viewer.current_colname()
            if current_col:
                if current_col.isidentifier():
                    default_text = f"c.{current_col}"
                else:
                    safe_name = current_col.replace('"', '\\"')
                    default_text = f'c["{safe_name}"]'

        def accept(buff):
            raw_text = buff.text
            text = raw_text.strip()
            if text.lower() == "cancel":
                screen.viewer.status_message = "filter canceled"
                screen._remove_modal(event.app)
                screen.refresh()
                return True

            try:
                args = [text] if text else []
                result = screen._runtime.invoke(
                    "filter_expr",
                    args=args,
                    source="tui",
                    context_mutator=screen._mutate_context,
                    propagate=(FilterError,),
                )
            except FilterError as err:
                screen._open_error_modal(
                    event,
                    "Filter Error",
                    str(err),
                    retry=lambda ev: self.open_filter_modal(ev, initial_text=raw_text),
                )
            except Exception as exc:
                screen._open_error_modal(
                    event,
                    "Unexpected Error",
                    str(exc),
                    retry=lambda ev: self.open_filter_modal(ev, initial_text=raw_text),
                )
            else:
                dispatch = screen._finalise_runtime_result(result)
                status_error = screen._status_error_message(("filter error",))
                error_message = result.message or status_error
                if error_message:
                    screen._open_error_modal(
                        event,
                        "Filter Error",
                        error_message,
                        retry=lambda ev: self.open_filter_modal(ev, initial_text=raw_text),
                    )
                    return True
                if dispatch is not None:
                    self.record_expr_filter(text)
                screen.viewer.status_message = None
                screen._remove_modal(event.app)
                screen.refresh()
            return True

        filter_field = TextArea(
            text=default_text,
            multiline=True,
            height=4,
            accept_handler=accept,
            history=None,
            completer=ColumnNameCompleter(screen.viewer.columns),
            complete_while_typing=True,
        )
        filter_field.buffer.cursor_position = len(default_text)

        tui_modals.bind_enter_to_accept(filter_field)
        body = tui_modals.build_prompt_body(prompt_text, filter_field)
        dialog = Dialog(title=title, body=body, buttons=[])
        screen._display_modal(
            event.app,
            dialog,
            focus=filter_field,
            context_type="expr_filter",
            payload={"field": filter_field},
            width=80,
        )

    def open_filter_modal_with_text(self, event, text: str) -> None:
        self.open_filter_modal(event, initial_text=text)

    def open_transform_modal(self, event, *, initial_text: str | None = None) -> None:
        screen = self._screen
        if screen.viewer.plan_controller.current_plan() is None:
            screen.viewer.status_message = "transform unsupported for this view"
            screen.refresh()
            return
        title = "Transform"
        existing_text = getattr(screen.viewer, "_transform_text", None)
        prompt_text = (
            "Polars LazyFrame transform (lf -> LazyFrame) - Enter: replace view"
            if existing_text
            else "Polars LazyFrame transform (lf -> LazyFrame) - Enter: derived view"
        )
        history = InMemoryHistory()
        for item in self._transform_history:
            history.append_string(item)

        default_text = initial_text or "lf"
        if initial_text is None and existing_text:
            default_text = existing_text
        elif initial_text is None:
            current_col = None
            with suppress(Exception):
                current_col = screen.viewer.current_colname()
            if current_col:
                if current_col.isidentifier():
                    default_text = f"lf.with_columns(c.{current_col})"
                else:
                    safe_name = current_col.replace('"', '\\"')
                    default_text = f'lf.with_columns(c["{safe_name}"])'

        def accept(buff):
            raw_text = buff.text
            text = raw_text.strip()
            if text.lower() == "cancel":
                screen.viewer.status_message = "transform canceled"
                screen._remove_modal(event.app)
                screen.refresh()
                return True

            try:
                result = screen._runtime.invoke(
                    "transform_expr",
                    args=[text],
                    source="tui",
                    context_mutator=screen._mutate_context,
                    propagate=(TransformError,),
                )
            except TransformError as err:
                screen._open_error_modal(
                    event,
                    "Transform Error",
                    str(err),
                    retry=lambda ev: self.open_transform_modal(ev, initial_text=raw_text),
                )
            except Exception as exc:
                screen._open_error_modal(
                    event,
                    "Unexpected Error",
                    str(exc),
                    retry=lambda ev: self.open_transform_modal(ev, initial_text=raw_text),
                )
            else:
                dispatch = screen._finalise_runtime_result(result)
                error_message = result.message or screen._status_error_message(("transform error",))
                if error_message:
                    screen._open_error_modal(
                        event,
                        "Transform Error",
                        error_message,
                        retry=lambda ev: self.open_transform_modal(ev, initial_text=raw_text),
                    )
                    return True
                if dispatch is not None:
                    self.record_transform(text)
                screen.viewer.status_message = None
                screen._remove_modal(event.app)
                screen.refresh()
            return True

        transform_field = TextArea(
            text=default_text,
            multiline=True,
            height=6,
            accept_handler=accept,
            history=history,
            completer=ColumnNameCompleter(screen.viewer.columns),
            complete_while_typing=True,
        )
        transform_field.buffer.cursor_position = len(default_text)

        tui_modals.bind_enter_to_accept(transform_field)
        body = tui_modals.build_prompt_body(prompt_text, transform_field)
        dialog = Dialog(title=title, body=body, buttons=[])
        screen._display_modal(
            event.app,
            dialog,
            focus=transform_field,
            context_type="transform",
            payload={"field": transform_field},
            width=80,
        )

    def open_sql_filter_modal(self, event, *, initial_text: str | None = None) -> None:
        screen = self._screen
        title = "SQL Filter"
        prompt_text = "SQL WHERE clause (column names only) - Enter: replace existing"
        default_text = initial_text or ""
        if not default_text:
            current_col = None
            with suppress(Exception):
                current_col = screen.viewer.current_colname()
            if current_col:
                if current_col.isidentifier():
                    default_text = current_col
                else:
                    default_text = ColumnNameCompleter._quote_identifier(current_col)

        def accept(buff):
            raw_text = buff.text
            text = raw_text.strip()
            if text.lower() == "cancel":
                screen.viewer.status_message = "SQL filter canceled"
                screen._remove_modal(event.app)
                screen.refresh()
                return True

            args = [text] if text else []
            try:
                result = screen._runtime.invoke(
                    "filter_sql",
                    args=args,
                    source="tui",
                    context_mutator=screen._mutate_context,
                    propagate=(FilterError,),
                )
            except Exception as exc:
                screen._open_error_modal(
                    event,
                    "SQL Filter Error",
                    str(exc),
                    retry=lambda ev: self.open_sql_filter_modal(ev, initial_text=raw_text),
                )
            else:
                dispatch = screen._finalise_runtime_result(result)
                status_error = screen._status_error_message(("sql filter error",))
                error_message = result.message or status_error
                if error_message:
                    screen._open_error_modal(
                        event,
                        "SQL Filter Error",
                        error_message,
                        retry=lambda ev: self.open_sql_filter_modal(ev, initial_text=raw_text),
                    )
                    return True
                if text and dispatch is not None:
                    self.record_sql_filter(text)
                screen.viewer.status_message = None
                screen._remove_modal(event.app)
                screen.refresh()
            return True

        filter_field = TextArea(
            text=default_text,
            multiline=True,
            height=4,
            accept_handler=accept,
            history=None,
            completer=ColumnNameCompleter(screen.viewer.columns, mode="sql"),
            complete_while_typing=True,
        )
        filter_field.buffer.cursor_position = len(default_text)

        tui_modals.bind_enter_to_accept(filter_field)
        body = tui_modals.build_prompt_body(prompt_text, filter_field)
        dialog = Dialog(title=title, body=body, buttons=[])
        screen._display_modal(
            event.app,
            dialog,
            focus=filter_field,
            context_type="sql_filter",
            payload={"field": filter_field},
            width=80,
        )

    def open_sql_filter_modal_with_text(self, event, text: str) -> None:
        self.open_sql_filter_modal(event, initial_text=text)

    def open_command_modal(self, event) -> None:
        screen = self._screen
        history = InMemoryHistory()
        for item in self._command_history:
            history.append_string(item)

        def accept(buff):
            raw_text = buff.text
            command_text = raw_text.strip()

            if not command_text:
                screen._remove_modal(event.app)
                screen.refresh()
                return True

            lowered = command_text.split(None, 1)[0].lower()
            if lowered in {"transform", "transform_expr"}:
                screen.viewer.status_message = "transform is available via Shift+E modal"
                screen._remove_modal(event.app)
                screen.refresh()
                return True

            result = screen._runtime.dispatch_raw(
                command_text,
                source="tui",
                context_mutator=screen._mutate_context,
            )
            dispatch = screen._finalise_runtime_result(result)
            if dispatch is not None:
                self.record_command(command_text)
            screen._remove_modal(event.app)
            screen.refresh()
            return True

        command_field = TextArea(
            text="",
            multiline=False,
            accept_handler=accept,
            history=history,
            completer=FilesystemPathCompleter(screen._path_completion_base_dir),
            complete_while_typing=True,
        )
        command_field.buffer.cursor_position = 0

        examples = []
        for spec in screen.commands.iter_specs():
            hints = spec.ui_hints or {}
            example = hints.get("example") if isinstance(hints, dict) else hints.get("example")
            if example:
                examples.append(str(example))
        unique_examples: list[str] = []
        for example in examples:
            if example not in unique_examples:
                unique_examples.append(example)
        prompt = (
            "Command:"
            if not unique_examples
            else f"Command (e.g. {', '.join(unique_examples[:3])}):"
        )

        body = tui_modals.build_prompt_body(prompt, command_field)
        dialog = Dialog(title="Command", body=body, buttons=[])
        screen._display_modal(
            event.app,
            dialog,
            focus=command_field,
            context_type="command",
            payload={"field": command_field},
            width=60,
        )

    def open_shell_modal(self, event) -> None:
        screen = self._screen
        if isinstance(screen.viewer.ui_hooks, NullViewerUIHooks):
            screen.viewer.status_message = "shell commands unavailable in headless mode"
            screen.refresh()
            return

        history = InMemoryHistory()
        for item in self._shell_history:
            history.append_string(item)

        def accept(buff):
            raw_text = buff.text
            command_text = raw_text.strip()

            if not command_text:
                screen._remove_modal(event.app)
                screen.refresh()
                return True

            self.record_shell_command(command_text)
            screen._remove_modal(event.app)
            screen.refresh()

            async def _run_shell() -> None:
                try:
                    exit_code = await run_in_terminal(
                        lambda: self._run_shell_command(command_text),
                        in_executor=True,
                    )
                except Exception as exc:
                    screen.viewer.status_message = f"shell error: {exc}"[:120]
                    screen.refresh()
                    return

                screen.viewer.status_message = f"shell exit {exit_code}"
                screen.refresh()

            if event.app is not None:
                event.app.create_background_task(_run_shell())
            else:
                exit_code = self._run_shell_command(command_text)
                screen.viewer.status_message = f"shell exit {exit_code}"
                screen.refresh()
            return True

        command_field = TextArea(
            text="",
            multiline=False,
            accept_handler=accept,
            history=history,
            completer=FilesystemPathCompleter(screen._path_completion_base_dir),
            complete_while_typing=True,
        )
        command_field.buffer.cursor_position = 0
        field_kb = KeyBindings()

        @field_kb.add("tab")
        def _trigger_completion(event) -> None:
            buff = event.current_buffer
            if buff.complete_state is None:
                buff.start_completion(select_first=True)
            buff.complete_next()

        tui_modals.merge_text_area_key_bindings(command_field, field_kb)

        body = tui_modals.build_prompt_body("! command:", command_field)
        dialog = Dialog(title="Shell", body=body, buttons=[])
        screen._display_modal(
            event.app,
            dialog,
            focus=command_field,
            context_type="shell",
            payload={"field": command_field},
            width=60,
        )

    def _run_shell_command(self, command_text: str) -> int:
        screen = self._screen
        shell = os.environ.get("SHELL") or "/bin/sh"
        cwd: Path | None = None
        with suppress(Exception):
            base_dir = screen._path_completion_base_dir()
            if base_dir is not None and base_dir.exists():
                cwd = base_dir
        exit_code = 1
        try:
            result = subprocess.run(
                [shell, "-lc", command_text],
                check=False,
                stdin=subprocess.DEVNULL,
                cwd=cwd,
            )
            exit_code = result.returncode
        except FileNotFoundError:
            result = subprocess.run(
                ["/bin/sh", "-lc", command_text],
                check=False,
                stdin=subprocess.DEVNULL,
                cwd=cwd,
            )
            exit_code = result.returncode
        except Exception as exc:
            print(f"shell error: {exc}")

        with suppress(EOFError):
            input("Press Enter to return to Pulka...")
        return exit_code

    def open_row_search_modal(self, event) -> None:
        screen = self._screen

        def accept(buff):
            text = buff.text.strip()
            if not text:
                screen._remove_modal(event.app)
                screen.refresh()
                return True

            if text.lower() == "cancel":
                screen.viewer.status_message = "row selection canceled"
                screen._remove_modal(event.app)
                screen.refresh()
                return True

            try:
                columns = ()
                try:
                    columns = (screen.viewer.columns[screen.viewer.cur_col],)
                except Exception:
                    columns = screen.viewer.columns[:1]
                screen.viewer.select_rows_containing(text, columns=columns)
                self.record_row_search(text)
            except Exception as exc:
                screen.viewer.status_message = f"Row selection error: {exc}"
            screen._remove_modal(event.app)
            screen.refresh()
            return True

        history = InMemoryHistory()
        for item in self._row_search_history:
            history.append_string(item)

        search_field = TextArea(
            text="",
            multiline=False,
            accept_handler=accept,
            history=history,
        )
        search_field.buffer.cursor_position = 0

        body = tui_modals.build_prompt_body(
            "Substring (current column, case-insensitive):",
            search_field,
        )
        dialog = Dialog(title="Select Rows (current column)", body=body, buttons=[])
        screen._display_modal(
            event.app,
            dialog,
            focus=search_field,
            context_type="row_search",
            payload={"field": search_field},
            width=60,
        )

    def open_filter_contains_modal(self, event) -> None:
        screen = self._screen

        def accept(buff):
            cleaned_text = buff.text.strip()
            if not cleaned_text:
                screen._remove_modal(event.app)
                screen.refresh()
                return True

            if cleaned_text.lower() == "cancel":
                screen.viewer.status_message = "filter canceled"
                screen._remove_modal(event.app)
                screen.refresh()
                return True

            try:
                sheet = screen.viewer.sheet
                traits = resolve_sheet_traits(sheet)
                if traits.is_file_browser:
                    setter = getattr(sheet, "set_contains_filter", None)
                    if callable(setter):
                        changed = setter(cleaned_text)
                        if changed:
                            self.record_filter_contains(cleaned_text)
                            controller = getattr(screen, "_file_browser_controller", None)
                            if controller is not None:
                                controller.refresh_sheet(sheet, screen.viewer)
                            else:
                                screen.viewer.status_message = file_browser_status_text(sheet)
                            screen.viewer.clamp()
                        else:
                            screen.viewer.status_message = "filter unchanged"
                    else:
                        screen.viewer.status_message = "filtering not supported"
                else:
                    screen.viewer.append_filter_for_contains_text(cleaned_text)
                    self.record_filter_contains(cleaned_text)
            except Exception as exc:
                screen.viewer.status_message = f"filter error: {exc}"
            screen._remove_modal(event.app)
            screen.refresh()
            return True

        history = InMemoryHistory()
        for item in self._filter_contains_history:
            history.append_string(item)

        search_field = TextArea(
            text="",
            multiline=False,
            accept_handler=accept,
            history=history,
        )
        search_field.buffer.cursor_position = 0

        body = tui_modals.build_prompt_body(
            "Substring (current column, case-insensitive):",
            search_field,
        )
        dialog = Dialog(title="Filter Rows (current column)", body=body, buttons=[])
        screen._display_modal(
            event.app,
            dialog,
            focus=search_field,
            context_type="filter_contains",
            payload={"field": search_field},
            width=60,
        )

    def open_search_modal(self, event) -> None:
        screen = self._screen

        def accept(buff):
            text = buff.text.strip()
            if text.lower() == "cancel":
                screen.viewer.status_message = "search canceled"
                screen._remove_modal(event.app)
                screen.refresh()
                return True

            # Apply search to the viewer
            try:
                screen.viewer.set_search(text)
                current = screen.viewer.search_text
                screen._clear_column_search()
                self.record_search(text)
                if current:
                    screen.viewer.search(forward=True, include_current=True)
                screen._remove_modal(event.app)
                screen.refresh()
            except Exception as exc:
                screen.viewer.status_message = f"Search error: {exc}"
            return True

        history = InMemoryHistory()
        for item in self._search_history:
            history.append_string(item)

        search_field = TextArea(
            text="",
            multiline=False,
            accept_handler=accept,
            history=history,
        )
        search_field.buffer.cursor_position = 0

        body = tui_modals.build_prompt_body(
            "Substring (current column, case-insensitive):",
            search_field,
        )
        dialog = Dialog(title="Search", body=body, buttons=[])
        screen._display_modal(
            event.app,
            dialog,
            focus=search_field,
            context_type="search",
            payload={"field": search_field},
            width=60,
        )

    def open_column_search_modal(self, event) -> None:
        """Open the column search modal with history and tab completion."""
        screen = self._screen

        def accept(buff):
            raw_text = buff.text
            query = raw_text.strip()

            if not query or query.lower() == "cancel":
                screen.viewer.status_message = "column search canceled"
                screen._clear_column_search()
                screen._remove_modal(event.app)
                screen.refresh()
                return True

            screen._remove_modal(event.app)
            success = screen._apply_column_search(query)
            self.record_column_search(query)
            if not success:
                screen._clear_column_search()
            screen.refresh()
            return True

        history = InMemoryHistory()
        for item in self._col_search_history:
            history.append_string(item)

        search_field = TextArea(
            text="",
            multiline=False,
            accept_handler=accept,
            history=history,
            completer=ColumnNameCompleter(screen.viewer.columns, mode="plain"),
            complete_while_typing=True,
        )
        search_field.buffer.cursor_position = 0

        body = tui_modals.build_prompt_body("Column name (prefix or substring):", search_field)
        dialog = Dialog(title="Column Search", body=body, buttons=[])
        screen._display_modal(
            event.app,
            dialog,
            focus=search_field,
            context_type="column_search",
            payload={"field": search_field},
            width=60,
        )

    def record_search(self, text: str) -> None:
        text = text.strip()
        if not text:
            return
        with suppress(ValueError):
            self._search_history.remove(text)
        self._search_history.append(text)
        if len(self._search_history) > self._history_max_size:
            del self._search_history[0]

    def record_row_search(self, text: str) -> None:
        cleaned = text.strip()
        if not cleaned:
            return
        with suppress(ValueError):
            self._row_search_history.remove(cleaned)
        self._row_search_history.append(cleaned)
        if len(self._row_search_history) > self._history_max_size:
            del self._row_search_history[0]

    def record_filter_contains(self, text: str) -> None:
        cleaned = text.strip()
        if not cleaned:
            return
        with suppress(ValueError):
            self._filter_contains_history.remove(cleaned)
        self._filter_contains_history.append(cleaned)
        if len(self._filter_contains_history) > self._history_max_size:
            del self._filter_contains_history[0]

    def record_command(self, text: str) -> None:
        cleaned = text.strip()
        if not cleaned:
            return
        with suppress(ValueError):
            self._command_history.remove(cleaned)
        self._command_history.append(cleaned)
        if len(self._command_history) > self._history_max_size:
            del self._command_history[0]

    def record_shell_command(self, text: str) -> None:
        cleaned = text.strip()
        if not cleaned:
            return
        with suppress(ValueError):
            self._shell_history.remove(cleaned)
        self._shell_history.append(cleaned)
        if len(self._shell_history) > self._history_max_size:
            del self._shell_history[0]

    def record_column_search(self, text: str) -> None:
        cleaned = text.strip()
        if not cleaned:
            return
        with suppress(ValueError):
            self._col_search_history.remove(cleaned)
        self._col_search_history.append(cleaned)
        if len(self._col_search_history) > self._history_max_size:
            del self._col_search_history[0]

    def record_expr_filter(self, text: str) -> None:
        cleaned = text.strip()
        if not cleaned:
            return
        with suppress(ValueError):
            self._expr_filter_history.remove(cleaned)
        self._expr_filter_history.append(cleaned)
        if len(self._expr_filter_history) > self._history_max_size:
            del self._expr_filter_history[0]

    def record_sql_filter(self, text: str) -> None:
        cleaned = text.strip()
        if not cleaned:
            return
        with suppress(ValueError):
            self._sql_filter_history.remove(cleaned)
        self._sql_filter_history.append(cleaned)
        if len(self._sql_filter_history) > self._history_max_size:
            del self._sql_filter_history[0]

    def record_transform(self, text: str) -> None:
        cleaned = text.strip()
        if not cleaned:
            return
        with suppress(ValueError):
            self._transform_history.remove(cleaned)
        self._transform_history.append(cleaned)
        if len(self._transform_history) > self._history_max_size:
            del self._transform_history[0]



"""Clipboard region selection handling for the TUI."""

from __future__ import annotations

from collections.abc import Callable
from contextlib import suppress
from dataclasses import dataclass

from ..core.viewer import Viewer


@dataclass(slots=True)
class ClipboardRegionState:
    """Track the starting cell for a clipboard region selection."""

    row_id: int | None = None
    col_index: int | None = None
    awaiting_start: bool = True
    format: str = "tsv"


class ClipboardRegionController:
    """Manage clipboard region selection and copying."""

    def __init__(
        self,
        *,
        get_viewer: Callable[[], Viewer | None],
        refresh: Callable[[], None],
        reset_pending_moves: Callable[[], None],
        copy_to_clipboard: Callable[[str, str, str], None],
    ) -> None:
        self._get_viewer = get_viewer
        self._refresh = refresh
        self._reset_pending_moves = reset_pending_moves
        self._copy_to_clipboard = copy_to_clipboard
        self._state: ClipboardRegionState | None = None

    def is_active(self) -> bool:
        return self._state is not None

    def awaiting_start(self) -> bool:
        return bool(self._state and self._state.awaiting_start)

    def start(self, *, format_name: str = "tsv") -> None:
        viewer = self._get_viewer()
        if viewer is None:
            return
        with suppress(Exception):
            viewer._ui_state.pop("clipboard_region_anchor", None)
        self._reset_pending_moves()
        columns = list(getattr(viewer, "visible_cols", []) or [])
        row_positions = list(getattr(viewer, "visible_row_positions", []) or [])
        if not columns or not row_positions:
            viewer.status_message = "no visible cells to copy"
            return
        self._state = ClipboardRegionState(format=format_name)
        viewer.status_message = "select first corner cell, Enter to set (Esc to cancel)"

    def start_markdown(self) -> None:
        self.start(format_name="markdown")

    def start_ascii(self) -> None:
        self.start(format_name="ascii")

    def start_unicode(self) -> None:
        self.start(format_name="unicode")

    def set_start(self) -> None:
        state = self._state
        viewer = self._get_viewer()
        if state is None or viewer is None:
            return
        indices = self._resolve_viewport_cursor_indices(viewer)
        if indices is None:
            viewer.status_message = "no visible cells to copy"
            return
        row_index, col_visible_index = indices
        row_positions = list(getattr(viewer, "visible_row_positions", []) or [])
        columns = list(getattr(viewer, "visible_cols", []) or [])
        if not row_positions or not columns:
            viewer.status_message = "no visible cells to copy"
            return
        if row_index < 0 or row_index >= len(row_positions):
            viewer.status_message = "no visible cells to copy"
            return
        if col_visible_index < 0 or col_visible_index >= len(columns):
            viewer.status_message = "no visible cells to copy"
            return

        row_id = row_positions[row_index]
        col_name = columns[col_visible_index]
        col_index = None
        all_cols = list(getattr(viewer, "columns", []) or [])
        with suppress(ValueError):
            col_index = all_cols.index(col_name)
        if col_index is None:
            viewer.status_message = "no visible cells to copy"
            return

        state.row_id = row_id
        state.col_index = col_index
        state.awaiting_start = False
        with suppress(Exception):
            viewer._ui_state["clipboard_region_anchor"] = (row_id, col_index)
        viewer.status_message = "select second corner cell, Enter to copy (Esc to cancel)"

    def set_start_at(self, row_id: int, col_name: str) -> None:
        state = self._state
        viewer = self._get_viewer()
        if state is None or viewer is None:
            return
        row_positions = list(getattr(viewer, "visible_row_positions", []) or [])
        columns = list(getattr(viewer, "visible_cols", []) or [])
        if not row_positions or not columns:
            viewer.status_message = "no visible cells to copy"
            return
        if row_id not in row_positions:
            viewer.status_message = "selection outside viewport"
            return
        if col_name not in columns:
            viewer.status_message = "selection outside viewport"
            return

        all_cols = list(getattr(viewer, "columns", []) or [])
        try:
            col_index = all_cols.index(col_name)
        except ValueError:
            viewer.status_message = "selection outside viewport"
            return

        state.row_id = row_id
        state.col_index = col_index
        state.awaiting_start = False
        with suppress(Exception):
            viewer._ui_state["clipboard_region_anchor"] = (row_id, col_index)
        viewer.status_message = "select second corner cell, Enter to copy (Esc to cancel)"

    def cancel(self) -> None:
        viewer = self._get_viewer()
        if self._state is None or viewer is None:
            return
        with suppress(Exception):
            viewer._ui_state.pop("clipboard_region_anchor", None)
        self._state = None
        self._reset_pending_moves()
        viewer.status_message = "clipboard region canceled"

    def finalize(self) -> None:
        state = self._state
        viewer = self._get_viewer()
        if state is None or viewer is None:
            return
        with suppress(Exception):
            viewer._ui_state.pop("clipboard_region_anchor", None)
        self._state = None
        self._reset_pending_moves()
        if state.awaiting_start or state.row_id is None or state.col_index is None:
            viewer.status_message = "select first corner cell first"
            return
        indices = self._resolve_viewport_cursor_indices(viewer)
        if indices is None:
            viewer.status_message = "no visible cells to copy"
            return
        self._copy_clipboard_region_to_clipboard(viewer, state, indices)

    def finalize_at(self, end: tuple[int, int]) -> None:
        state = self._state
        viewer = self._get_viewer()
        if state is None or viewer is None:
            return
        with suppress(Exception):
            viewer._ui_state.pop("clipboard_region_anchor", None)
        self._state = None
        self._reset_pending_moves()
        if state.awaiting_start or state.row_id is None or state.col_index is None:
            viewer.status_message = "select first corner cell first"
            return
        self._copy_clipboard_region_to_clipboard(viewer, state, end)

    def on_cell_click(self, row_id: int, col_name: str) -> None:
        state = self._state
        viewer = self._get_viewer()
        if state is None or viewer is None:
            return
        row_positions = list(getattr(viewer, "visible_row_positions", []) or [])
        columns = list(getattr(viewer, "visible_cols", []) or [])
        if not row_positions or not columns:
            viewer.status_message = "no visible cells to copy"
            return
        if row_id not in row_positions:
            viewer.status_message = "selection outside viewport"
            return
        if col_name not in columns:
            viewer.status_message = "selection outside viewport"
            return
        row_index = row_positions.index(row_id)
        col_visible_index = columns.index(col_name)
        if state.awaiting_start:
            self.set_start_at(row_id, col_name)
        else:
            self.finalize_at((row_index, col_visible_index))

    def _resolve_viewport_cursor_indices(self, viewer: Viewer) -> tuple[int, int] | None:
        columns = list(getattr(viewer, "visible_cols", []) or [])
        if not columns:
            return None
        try:
            current_col = viewer.current_colname()
        except Exception:
            current_col = None
        if not current_col:
            cur_col = getattr(viewer, "cur_col", None)
            if isinstance(cur_col, int):
                all_cols = getattr(viewer, "columns", [])
                if 0 <= cur_col < len(all_cols):
                    current_col = all_cols[cur_col]
        if not current_col:
            return None
        try:
            col_index = columns.index(current_col)
        except ValueError:
            return None

        row_positions = list(getattr(viewer, "visible_row_positions", []) or [])
        if not row_positions:
            return None
        cur_row = getattr(viewer, "cur_row", None)
        if not isinstance(cur_row, int):
            return None
        try:
            row_index = row_positions.index(cur_row)
        except ValueError:
            row0 = getattr(viewer, "row0", 0)
            row_index = max(0, min(len(row_positions) - 1, cur_row - row0))
        return row_index, col_index

    @staticmethod
    def _sanitize_clipboard_cell(value: object) -> str:
        text = "" if value is None else str(value)
        return text.replace("\t", " ").replace("\n", " ").replace("\r", " ")

    @classmethod
    def _sanitize_markdown_cell(cls, value: object) -> str:
        text = cls._sanitize_clipboard_cell(value)
        return text.replace("|", "\\|")

    @classmethod
    def _sanitize_ascii_cell(cls, value: object) -> str:
        text = cls._sanitize_clipboard_cell(value)
        return text.replace("|", "\\|")

    def _copy_clipboard_region_to_clipboard(
        self, viewer: Viewer, start: ClipboardRegionState, _end: tuple[int, int]
    ) -> None:
        columns = list(getattr(viewer, "visible_cols", []) or [])
        row_positions = list(getattr(viewer, "visible_row_positions", []) or [])
        if not columns or not row_positions:
            viewer.status_message = "no visible data to copy"
            return

        if start.row_id is None or start.col_index is None:
            viewer.status_message = "select first corner cell first"
            return
        end_row_index, end_col_visible_index = _end
        if end_row_index < 0 or end_row_index >= len(row_positions):
            viewer.status_message = "selection outside viewport"
            return
        if end_col_visible_index < 0 or end_col_visible_index >= len(columns):
            viewer.status_message = "selection outside viewport"
            return

        end_row_id = row_positions[end_row_index]
        end_col_name = columns[end_col_visible_index]
        all_cols = list(getattr(viewer, "columns", []) or [])
        end_col_index: int | None = None
        with suppress(ValueError):
            end_col_index = all_cols.index(end_col_name)
        if end_col_index is None:
            viewer.status_message = "selection outside viewport"
            return

        row_lo_id = min(start.row_id, end_row_id)
        row_hi_id = max(start.row_id, end_row_id)
        col_lo = min(start.col_index, end_col_index)
        col_hi = max(start.col_index, end_col_index)

        visible_columns_fn = getattr(viewer, "visible_columns", None)
        if callable(visible_columns_fn):
            visible_columns = list(visible_columns_fn() or [])
        else:
            hidden_cols = set(getattr(viewer, "_hidden_cols", set()) or set())
            visible_columns = [name for name in all_cols if name not in hidden_cols]

        visible_set = set(visible_columns)
        selected_columns = [
            name
            for idx, name in enumerate(all_cols)
            if col_lo <= idx <= col_hi and name in visible_set
        ]
        if not selected_columns:
            viewer.status_message = "no columns to copy"
            return

        selection_active = False
        has_selection = getattr(viewer, "has_active_selection", None)
        if callable(has_selection):
            selection_active = has_selection()
        else:
            selection_active = bool(
                getattr(viewer, "_selected_row_ids", None)
                or getattr(viewer, "_selection_filter_expr", None)
                or getattr(viewer, "_value_selection_filter", None)
            )

        selection_lookup = set(getattr(viewer, "_selected_row_ids", set()) or set())
        selection_expr = None
        if selection_active:
            selection_expr = getattr(viewer, "_selection_filter_expr", None)
            if not selection_expr:
                value_expr_fn = getattr(viewer, "_value_selection_filter_expr", None)
                if callable(value_expr_fn):
                    try:
                        selection_expr = value_expr_fn()
                    except Exception:
                        selection_expr = None

        row_start = min(row_lo_id, row_hi_id)
        row_end = max(row_lo_id, row_hi_id)
        if row_end < row_start:
            viewer.status_message = "no rows to copy"
            return

        fetch_columns = list(selected_columns)
        if selection_expr:
            fetch_columns = list(getattr(viewer, "columns", []) or [])
            if not fetch_columns:
                viewer.status_message = "no columns to copy"
                return

        format_name = start.format if start.format else "tsv"
        sanitize = self._sanitize_clipboard_cell
        if format_name in {"markdown", "ascii", "unicode"}:
            sanitize = (
                self._sanitize_markdown_cell
                if format_name == "markdown"
                else self._sanitize_ascii_cell
            )
            header_cells = [sanitize(name) for name in selected_columns]
            rows: list[list[str]] = []
        else:
            header_line = "\t".join(sanitize(name) for name in selected_columns)
            lines = [header_line]
        rows_copied = 0

        row_provider = getattr(viewer, "row_provider", None)
        if row_provider is None:
            viewer.status_message = "clipboard copy unavailable"
            return

        plan = getattr(viewer, "_current_plan", None)
        plan_value = plan() if callable(plan) else None

        page_size = getattr(row_provider, "page_size", 512)
        try:
            chunk_size = max(1, int(page_size))
        except (TypeError, ValueError):
            chunk_size = 512

        cursor = row_start
        while cursor <= row_end:
            fetch_count = min(chunk_size, row_end - cursor + 1)
            try:
                table_slice, _ = row_provider.get_slice(
                    plan_value, fetch_columns, cursor, fetch_count
                )
            except Exception as exc:
                viewer.status_message = f"clipboard copy failed: {exc}"[:120]
                return

            height = getattr(table_slice, "height", 0)
            if height <= 0:
                break

            base_row = table_slice.start_offset if table_slice.start_offset is not None else cursor
            row_positions = range(base_row, base_row + height)

            selection_matches = None
            if selection_expr:
                resolver = getattr(viewer, "_selection_matches_for_slice", None)
                if callable(resolver):
                    selection_matches = resolver(table_slice, row_positions, selection_expr)
                if selection_matches is None and not selection_lookup:
                    viewer.status_message = "selection unavailable for clipboard copy"
                    return

            id_resolver = getattr(viewer, "_row_identifier_for_slice", None)
            if selection_active and not callable(id_resolver):
                viewer.status_message = "selection unavailable for clipboard copy"
                return

            column_values = [table_slice.column(name).formatted(0) for name in selected_columns]
            for idx in range(height):
                if selection_active:
                    row_id = id_resolver(
                        table_slice,
                        idx,
                        row_positions=row_positions,
                        absolute_row=base_row + idx,
                    )
                    if row_id is None:
                        continue
                    if selection_lookup or selection_matches is not None:
                        if row_id not in selection_lookup and (
                            selection_matches is None or row_id not in selection_matches
                        ):
                            continue
                    else:
                        continue

                row_values = []
                for col_values in column_values:
                    try:
                        cell = col_values[idx]
                    except Exception:
                        cell = ""
                    row_values.append(sanitize(cell))
                if format_name in {"markdown", "ascii", "unicode"}:
                    rows.append(row_values)
                else:
                    lines.append("\t".join(row_values))
                rows_copied += 1

            next_cursor = base_row + height
            if next_cursor <= cursor:
                cursor += max(1, height)
            else:
                cursor = next_cursor

        if rows_copied <= 0:
            viewer.status_message = (
                "no selected rows in region" if selection_active else "no rows to copy"
            )
            return

        if format_name == "markdown":
            header_line = "| " + " | ".join(header_cells) + " |"
            sep_line = "| " + " | ".join("---" for _ in header_cells) + " |"
            row_lines = ["| " + " | ".join(row) + " |" for row in rows]
            payload = "\n".join([header_line, sep_line, *row_lines])
        elif format_name in {"ascii", "unicode"}:
            widths = [len(cell) for cell in header_cells]
            for row in rows:
                for idx, cell in enumerate(row):
                    widths[idx] = max(widths[idx], len(cell))

            def _pad_cells(cells: list[str]) -> list[str]:
                return [
                    f" {cell}{' ' * (width - len(cell))} "
                    for cell, width in zip(cells, widths, strict=True)
                ]

            if format_name == "unicode":
                top = "┌" + "┬".join("─" * (width + 2) for width in widths) + "┐"
                mid = "├" + "┼".join("─" * (width + 2) for width in widths) + "┤"
                bottom = "└" + "┴".join("─" * (width + 2) for width in widths) + "┘"
                vert = "│"
            else:
                top = "+" + "+".join("-" * (width + 2) for width in widths) + "+"
                mid = top
                bottom = top
                vert = "|"
            header = vert + vert.join(_pad_cells(header_cells)) + vert
            row_lines = [vert + vert.join(_pad_cells(row)) + vert for row in rows]
            payload = "\n".join([top, header, mid, *row_lines, bottom])
        else:
            payload = "\n".join(lines)
        self._copy_to_clipboard(
            payload,
            success_message=f"copied {rows_copied}x{len(selected_columns)} region to clipboard",
            failure_message="clipboard unavailable",
        )



"""Row slice provider that bridges viewer requests to engine adapters."""

from __future__ import annotations

import logging
from collections import OrderedDict
from collections.abc import Callable, Iterator, Sequence
from concurrent.futures import Future
from threading import RLock
from typing import TYPE_CHECKING, Any, cast

import polars as pl

from ..config.settings import CACHE_DEFAULTS, STREAMING_DEFAULTS
from .column_insight import CellPreview, summarize_value_preview
from .engine.contracts import TableColumn, TableSlice
from .engine.polars_adapter import table_slice_from_dataframe
from .errors import CompileError, MaterializeError, PulkaCoreError
from .interfaces import (
    EngineAdapterProtocol,
    JobRunnerProtocol,
    MaterializerProtocol,
    is_materializer_compatible,
)
from .jobs import JobRequest, JobResult
from .plan import QueryPlan, normalized_columns_key
from .row_provider_cache import RowCacheManager
from .row_provider_parquet import ParquetRowGroupIndexer
from .row_provider_sidecar import SidecarWindow, SidecarWindowManager
from .row_provider_streaming import RowStreamingPipeline
from .row_provider_types import (
    PlanContext,
    RowKey,
    SliceStatus,
    SliceStreamRequest,
    TableSliceChunk,
)
from .sheet import SHEET_FEATURE_SLICE, sheet_supports
from .strategy import Strategy, compile_strategy

LOGGER = logging.getLogger(__name__)


if TYPE_CHECKING:
    from .source_traits import SourceTraits


class RowProvider:
    """Serve row slices for a sheet, optionally prefetching upcoming ranges."""

    __slots__ = (
        "_engine_factory",
        "_columns_getter",
        "_fetcher",
        "_job_context",
        "_runner",
        "_lock",
        "_materializer",
        "_empty_result_factory",
        "_empty_template",
        "_cache",
        "_max_cache_cells",
        "_max_cache_entries",
        "_max_page_cache_cells",
        "_max_page_cache_entries",
        "_streaming_enabled",
        "_streaming_batch_rows",
        "_streaming_last_chunks",
        "_streaming_last_rows",
        "_streaming_last_cells",
        "_streaming_last_duration_ns",
        "_streaming_last_mode",
        "_source_traits_cache",
        "_strategy",
        "_strategy_cache",
        "_streaming_enabled_configured",
        "_streaming_batch_rows_configured",
        "_sidecar",
        "_page_size",
        "_row_id_column",
        "_parquet_indexer",
        "_streaming",
    )

    MAX_CACHE_CELLS = CACHE_DEFAULTS.row_provider_max_cells
    MAX_CACHE_ENTRIES = CACHE_DEFAULTS.row_provider_max_entries
    PAGE_SIZE = CACHE_DEFAULTS.row_provider_page_size
    STREAMING_ENABLED = STREAMING_DEFAULTS.enabled
    STREAMING_BATCH_ROWS = STREAMING_DEFAULTS.batch_rows
    SOURCE_TRAITS_CACHE_LIMIT = 8
    STRATEGY_CACHE_LIMIT = SOURCE_TRAITS_CACHE_LIMIT

    _max_cache_cells: int
    _max_cache_entries: int
    _max_page_cache_cells: int
    _max_page_cache_entries: int

    _LIMIT_ATTRS = {
        "_max_cache_cells",
        "_max_cache_entries",
        "_max_page_cache_cells",
        "_max_page_cache_entries",
    }

    def __init__(
        self,
        *,
        engine_factory: Callable[[], EngineAdapterProtocol] | None = None,
        columns_getter: Callable[[], Sequence[str]] | None = None,
        materializer: MaterializerProtocol | None = None,
        fetcher: Callable[[int, int, Sequence[str]], Any] | None = None,
        job_context: Callable[[], tuple[str, int, str]] | None = None,
        empty_result_factory: Callable[[], Any] | None = None,
        runner: JobRunnerProtocol,
        streaming_enabled: bool | None = None,
        streaming_batch_rows: int | None = None,
        row_id_column: str | None = None,
    ) -> None:
        if engine_factory is None:
            if fetcher is None:
                msg = "RowProvider requires either engine_factory or fetcher"
                raise ValueError(msg)
            self._fetcher: Callable[[int, int, Sequence[str]], Any] | None = fetcher
            self._engine_factory = None
            self._columns_getter = None
            self._materializer = None
        else:
            if columns_getter is None:
                msg = "columns_getter is required when engine_factory is provided"
                raise ValueError(msg)
            if materializer is None:
                msg = "materializer is required when engine_factory is provided"
                raise ValueError(msg)
            if not is_materializer_compatible(materializer):
                msg = "materializer must implement MaterializerProtocol"
                raise TypeError(msg)
            self._engine_factory = engine_factory
            self._columns_getter = columns_getter
            self._materializer = materializer
            self._fetcher = None

        if empty_result_factory is None:
            msg = "RowProvider requires an empty_result_factory"
            raise ValueError(msg)

        self._empty_result_factory = empty_result_factory
        self._job_context = job_context
        if runner is None:
            msg = "RowProvider requires a JobRunner instance"
            raise ValueError(msg)
        if not isinstance(runner, JobRunnerProtocol):
            msg = "runner must implement JobRunnerProtocol"
            raise TypeError(msg)
        self._runner = runner
        self._empty_template: TableSlice | None = None
        self._streaming_enabled_configured = streaming_enabled is not None
        self._streaming_enabled = (
            self.STREAMING_ENABLED if streaming_enabled is None else bool(streaming_enabled)
        )
        batch_default = self.STREAMING_BATCH_ROWS
        self._streaming_batch_rows_configured = (
            streaming_batch_rows is not None and streaming_batch_rows > 0
        )
        if streaming_batch_rows is not None and streaming_batch_rows > 0:
            batch_default = int(streaming_batch_rows)
        self._streaming_batch_rows = max(1, batch_default)
        self._streaming_last_chunks = 0
        self._streaming_last_rows = 0
        self._streaming_last_cells = 0
        self._streaming_last_duration_ns = 0
        self._streaming_last_mode = "init"
        self._lock = RLock()
        self._page_size = max(1, self.PAGE_SIZE)
        self._row_id_column = row_id_column
        object.__setattr__(self, "_max_cache_cells", self.MAX_CACHE_CELLS)
        object.__setattr__(self, "_max_cache_entries", self.MAX_CACHE_ENTRIES)
        object.__setattr__(self, "_max_page_cache_cells", self.MAX_CACHE_CELLS)
        object.__setattr__(self, "_max_page_cache_entries", self.MAX_CACHE_ENTRIES)
        self._cache = RowCacheManager(
            lock=self._lock,
            cell_counter=self._cell_count,
            max_cache_cells=self._max_cache_cells,
            max_cache_entries=self._max_cache_entries,
            max_page_cache_cells=self._max_page_cache_cells,
            max_page_cache_entries=self._max_page_cache_entries,
        )
        self._source_traits_cache: OrderedDict[str, SourceTraits] = OrderedDict()
        self._strategy: Strategy | None = None
        self._strategy_cache: OrderedDict[str, Strategy] = OrderedDict()
        self._sidecar = SidecarWindowManager(
            lock=self._lock,
            runner=self._runner,
            traits_getter=self.get_source_traits,
        )
        self._parquet_indexer = ParquetRowGroupIndexer(lock=self._lock)
        self._streaming = RowStreamingPipeline(self)

    def __setattr__(self, name: str, value: Any) -> None:
        if name in self._LIMIT_ATTRS:
            object.__setattr__(self, name, value)
            cache = getattr(self, "_cache", None)
            if cache is not None:
                cache.update_limits(
                    max_cache_cells=self._max_cache_cells,
                    max_cache_entries=self._max_cache_entries,
                    max_page_cache_cells=self._max_page_cache_cells,
                    max_page_cache_entries=self._max_page_cache_entries,
                )
            return
        super().__setattr__(name, value)

    # ------------------------------------------------------------------
    # Construction helpers
    # ------------------------------------------------------------------
    @classmethod
    def for_sheet(cls, sheet: Any, *, runner: JobRunnerProtocol) -> RowProvider:
        """Build a provider for ``sheet`` by introspecting available hooks."""

        config_factory = getattr(sheet, "row_provider_config", None)
        if callable(config_factory):
            config = dict(config_factory())
            return cls(runner=runner, **config)

        if sheet_supports(sheet, SHEET_FEATURE_SLICE):

            def fetcher(start: int, count: int, cols: Sequence[str]) -> Any:
                return sheet.fetch_slice(start, count, list(cols))

            def empty_result() -> Any:
                base_columns = list(getattr(sheet, "columns", []))
                try:
                    return sheet.fetch_slice(0, 0, base_columns)
                except Exception:
                    return sheet.fetch_slice(0, 0, [])

            job_ctx = getattr(sheet, "job_context", None)
            return cls(
                fetcher=fetcher,
                job_context=job_ctx,
                runner=runner,
                empty_result_factory=empty_result,
            )

        msg = "Sheet does not expose a supported row interface"
        raise TypeError(msg)

    @classmethod
    def for_plan_source(
        cls,
        *,
        engine_factory: Callable[[], EngineAdapterProtocol],
        columns_getter: Callable[[], Sequence[str]],
        job_context: Callable[[], tuple[str, int, str]] | None,
        materializer: MaterializerProtocol,
        empty_result_factory: Callable[[], Any],
        runner: JobRunnerProtocol,
        row_id_column: str | None = None,
    ) -> RowProvider:
        return cls(
            engine_factory=engine_factory,
            columns_getter=columns_getter,
            job_context=job_context,
            materializer=materializer,
            empty_result_factory=empty_result_factory,
            runner=runner,
            row_id_column=row_id_column,
        )

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------
    @property
    def page_size(self) -> int:
        return self._page_size

    def current_strategy(self) -> Strategy | None:
        """Return the active access strategy, if available."""

        return self._strategy

    def get_page_if_cached(
        self,
        plan: QueryPlan | None,
        columns: Sequence[str],
        start: int,
    ) -> tuple[TableSlice, SliceStatus] | None:
        """Return a cached page slice without triggering a fetch."""

        columns = tuple(columns)
        if start < 0:
            start = 0

        context = self._resolve_context(plan, columns)
        if context is None:
            empty = self._empty_slice(columns)
            return empty, SliceStatus.SCHEMA_MISMATCH

        count = self._page_size
        key = self._cache_key(context.plan_hash, start, count, context.fetch_columns)
        cached = self._cache.get_page_cache(key)
        if cached is None:
            return None

        slice_, status = self._finalize_slice(context, cached)
        return slice_, status

    def get_page(
        self,
        plan: QueryPlan | None,
        columns: Sequence[str],
        start: int,
    ) -> tuple[TableSlice, SliceStatus, bool]:
        """Return a fixed-size page starting at ``start``."""

        columns = tuple(columns)
        if start < 0:
            start = 0

        context = self._resolve_context(plan, columns)
        if context is None:
            empty = self._empty_slice(columns)
            return empty, SliceStatus.SCHEMA_MISMATCH, False

        count = self._page_size
        key = self._cache_key(context.plan_hash, start, count, context.fetch_columns)
        cache_hit = False
        cached = self._cache.get_page_cache(key)
        if cached is not None:
            cache_hit = True
        if cached is not None:
            slice_, status = self._finalize_slice(context, cached)
            return slice_, status, cache_hit

        if not context.fetch_columns:
            raw_slice, _ = self._fetch_slice(context, start, count)
            slice_, status = self._finalize_slice(context, raw_slice)
            self._cache.store_page_cache(key, raw_slice)
            return slice_, status, cache_hit

        self._ensure_strategy_for_context(context)
        raw_slice, _ = self._fetch_slice(context, start, count)
        self._cache.store_page_cache(key, raw_slice)
        slice_, status = self._finalize_slice(context, raw_slice)
        return slice_, status, cache_hit

    def get_slice(
        self,
        plan: QueryPlan | None,
        columns: Sequence[str],
        start: int,
        count: int,
    ) -> tuple[TableSlice, SliceStatus]:
        """Return ``count`` rows starting from ``start`` for ``columns``."""

        columns = tuple(columns)

        if count <= 0:
            empty = self._empty_slice(columns)
            status = SliceStatus.PARTIAL if columns else SliceStatus.OK
            return empty, status

        context = self._resolve_context(plan, columns)
        if context is None:
            empty = self._empty_slice(columns)
            return empty, SliceStatus.SCHEMA_MISMATCH

        if self._fetcher is None and context.fetch_columns:
            self.prepare_sidecar_window(context, start, count, record_progress=True)

        page_size = self._page_size
        page_start = max(0, (start // page_size) * page_size)
        end_row = max(start, start + count - 1)
        last_page_start = max(0, (end_row // page_size) * page_size)

        slices: list[TableSlice] = []
        status = SliceStatus.OK
        remaining = count
        cursor = start
        for page_offset in range(page_start, last_page_start + page_size, page_size):
            page_slice, page_status, _cache_hit = self.get_page(plan, columns, page_offset)
            status = self._merge_status(status, page_status)
            if page_slice.height <= 0:
                if remaining <= 0:
                    break
                cursor = max(cursor, page_offset + page_size)
                continue
            local_start = max(0, cursor - page_offset)
            local_len = min(remaining, max(0, page_slice.height - local_start))
            if local_len <= 0:
                cursor = max(cursor, page_offset + page_size)
                continue
            slices.append(page_slice.slice(local_start, local_len))
            cursor += local_len
            remaining -= local_len
            if remaining <= 0:
                break

        if not slices:
            empty = self._empty_slice(columns)
            return empty, status

        result = slices[0]
        for additional in slices[1:]:
            result = result.concat_vertical(additional)
        return result, status

    def iter_slice_stream(
        self,
        plan: QueryPlan | None,
        columns: Sequence[str],
        start: int,
        count: int | None,
        *,
        batch_rows: int | None = None,
    ) -> Iterator[TableSlice]:
        """Yield slices for ``[start, start + count)`` without assembling a composite."""

        columns = tuple(columns)

        if count is not None and count <= 0:
            return

        if start < 0:
            start = 0

        context = self._resolve_context(plan, columns)
        if context is None:
            yield self._empty_slice(columns)
            return

        if self._fetcher is not None:
            if count is None:
                chunk = self._page_size
                cursor = start
                while True:
                    slice_, _status = self.get_slice(plan, columns, cursor, chunk)
                    if slice_.height <= 0:
                        break
                    yield slice_
                    cursor = (
                        slice_.start_offset + slice_.height
                        if slice_.start_offset is not None
                        else cursor + slice_.height
                    )
                return
            slice_, _status = self.get_slice(plan, columns, start, count)
            yield slice_
            return

        prepared = self._prepare_materializer(context.plan)
        if prepared is None:
            yield self._empty_slice(columns)
            return

        materializer, physical_plan = prepared
        self._record_source_traits(context.plan, physical_plan)

        stream_attr = getattr(materializer, "collect_slice_stream", None)
        stream_iterator: Iterator[TableSlice] | None = None
        stream_kwargs: dict[str, object] = {"start": start, "columns": tuple(context.fetch_columns)}
        if count is not None:
            stream_kwargs["length"] = count
        if callable(stream_attr):
            if batch_rows is not None and batch_rows > 0:
                stream_kwargs["batch_rows"] = batch_rows
            try:
                stream_iterator = iter(stream_attr(physical_plan, **stream_kwargs))
            except TypeError:
                stream_kwargs.pop("batch_rows", None)
                try:
                    stream_iterator = iter(stream_attr(physical_plan, **stream_kwargs))
                except Exception:
                    stream_iterator = None
            except Exception:
                stream_iterator = None

        if stream_iterator is None:
            if count is None:
                chunk = self._page_size
                cursor = start
                while True:
                    slice_, _status = self.get_slice(plan, columns, cursor, chunk)
                    if slice_.height <= 0:
                        break
                    yield slice_
                    cursor = (
                        slice_.start_offset + slice_.height
                        if slice_.start_offset is not None
                        else cursor + slice_.height
                    )
                return
            slice_, _status = self.get_slice(plan, columns, start, count)
            yield slice_
            return

        total_rows = 0
        try:
            for raw_chunk in stream_iterator:
                chunk_slice = self._coerce_slice(raw_chunk, row_id_column=self._row_id_column)
                if chunk_slice.height <= 0:
                    continue
                remaining = None if count is None else max(0, count - total_rows)
                if remaining is not None and remaining <= 0:
                    break
                if remaining is not None and chunk_slice.height > remaining:
                    chunk_slice = chunk_slice.slice(0, remaining)
                if chunk_slice.height <= 0:
                    continue
                chunk_offset = start + total_rows
                total_rows += chunk_slice.height
                if chunk_slice.start_offset is None:
                    chunk_slice = TableSlice(
                        tuple(chunk_slice.columns),
                        chunk_slice.schema,
                        start_offset=chunk_offset,
                        row_ids=chunk_slice.row_ids,
                    )
                yield chunk_slice
                if count is not None and total_rows >= count:
                    break
        except Exception as exc:  # pragma: no cover - defensive
            msg = "Failed to stream row slices"
            raise MaterializeError(msg) from exc

    def prime_page_cache(
        self,
        plan: QueryPlan | None,
        columns: Sequence[str],
        raw_slice: TableSlice,
    ) -> None:
        """Seed the internal page cache with full pages covered by ``raw_slice``.

        This is a best-effort optimization used by navigation commands that
        already streamed through the target area and want the subsequent render
        to hit the row cache rather than triggering a cold seek.
        """

        if raw_slice.height <= 0:
            return
        if raw_slice.start_offset is None:
            return

        context = self._resolve_context(plan, tuple(columns))
        if context is None or not context.fetch_columns:
            return

        page_size = self._page_size
        slice_start = raw_slice.start_offset
        slice_end = slice_start + raw_slice.height
        first_full_page = ((slice_start + page_size - 1) // page_size) * page_size
        last_full_page = (slice_end // page_size) * page_size
        if first_full_page >= last_full_page:
            return

        for page_start in range(first_full_page, last_full_page, page_size):
            local_start = page_start - slice_start
            if local_start < 0:
                continue
            page_slice = raw_slice.slice(local_start, page_size)
            if page_slice.height != page_size:
                continue
            if page_slice.start_offset is None:
                page_slice = TableSlice(
                    tuple(page_slice.columns),
                    page_slice.schema,
                    start_offset=page_start,
                    row_ids=page_slice.row_ids,
                )
            key = self._cache_key(context.plan_hash, page_start, page_size, context.fetch_columns)
            self._cache.store_page_cache(key, page_slice)

    def get_slice_stream(self, request: SliceStreamRequest) -> Iterator[TableSliceChunk]:
        """Yield streaming chunks that resolve to the requested slice."""

        yield from self._streaming.get_slice_stream(request)

    def build_plan_compiler(self) -> EngineAdapterProtocol | None:
        """Return an engine adapter for validation when exposed by the sheet."""

        return self._engine_adapter()

    def prefetch(
        self,
        plan: QueryPlan | None,
        columns: Sequence[str],
        start: int,
        count: int,
    ) -> None:
        """Warm ``[start, start + count)`` in the background when possible."""

        if count <= 0:
            return

        if self._fetcher is not None:
            # Nothing clever to do for passthrough providers.
            return

        columns = tuple(columns)
        context = self._resolve_context(plan, columns)
        if context is None or not context.fetch_columns:
            return

        if context.sheet_id is None or context.generation is None:
            return

        page_size = self._page_size
        start_page = max(0, (start // page_size) * page_size)
        end_row = max(start, start + count - 1)
        end_page = max(0, (end_row // page_size) * page_size)
        for page_start in range(start_page, end_page + page_size, page_size):
            self.prefetch_page(plan, columns, page_start)

    def prefetch_page(
        self,
        plan: QueryPlan | None,
        columns: Sequence[str],
        start: int,
    ) -> Future[JobResult] | None:
        """Warm a single page starting at ``start`` in the background."""

        if start < 0:
            return None

        if self._fetcher is not None:
            return None

        columns = tuple(columns)
        context = self._resolve_context(plan, columns)
        if context is None or not context.fetch_columns:
            return None

        if context.sheet_id is None or context.generation is None:
            return None

        count = self._page_size
        key = self._cache_key(context.plan_hash, start, count, context.fetch_columns)
        scheduled = self._cache.reserve_page_prefetch(key)
        if not scheduled:
            return None

        tag_hash = context.plan_hash or "none"
        cols_sig = normalized_columns_key(context.fetch_columns)
        job_tag = f"rows:{tag_hash}:{start}:{count}:{cols_sig}"

        def _job(_: int) -> Any:
            slice_result, _prefetch_window = self._fetch_slice(
                context,
                start,
                count,
                record_progress=False,
            )
            return slice_result

        req = JobRequest(
            sheet_id=context.sheet_id,
            generation=context.generation,
            tag=job_tag,
            fn=_job,
            cache_result=False,
        )
        future = self._runner.enqueue(req)
        self._cache.register_prefetch_future(key, future)
        scheduled_count = self._cache.cache_metrics().get("prefetch_scheduled", 0)
        if scheduled and LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug(
                "row_provider.prefetch_schedule",
                extra={
                    "event": "row_prefetch_schedule",
                    "plan_hash": context.plan_hash,
                    "start": start,
                    "count": count,
                    "columns": cols_sig,
                    "scheduled": scheduled_count,
                },
            )

        def _store_result(fut: Any, *, row_key: RowKey = key) -> None:
            self._cache.clear_pending(row_key)

            try:
                result = fut.result()
            except Exception:
                return

            if getattr(result, "error", None) is not None:
                return

            value = getattr(result, "value", None)
            generation = getattr(result, "generation", None)
            if generation != context.generation:
                return

            if value is None:
                return

            self._cache.store_page_cache(row_key, value, prefetched=True)

        future.add_done_callback(_store_result)
        return future

    def current_cell_value(
        self,
        plan: QueryPlan | None,
        column: str,
        row: int,
        *,
        preview_chars: int = 160,
    ) -> CellPreview | None:
        """Return a lightweight preview for ``column`` at ``row``."""

        if row < 0:
            return None

        slice_, _status = self.get_slice(plan, (column,), row, 1)
        if slice_.height <= 0:
            return None

        try:
            table_column = slice_.column(column)
        except KeyError:
            return None

        values = table_column.values
        try:
            raw_value = values[0]
        except (IndexError, TypeError):
            return None
        display, truncated = summarize_value_preview(raw_value, max_chars=preview_chars)
        dtype = str(table_column.dtype) if table_column.dtype is not None else None
        absolute_row = slice_.start_offset
        if absolute_row is None:
            absolute_row = row

        return CellPreview(
            column=column,
            row=row,
            absolute_row=absolute_row,
            dtype=dtype,
            raw_value=raw_value,
            display=display,
            truncated=truncated,
        )

    def get_source_traits(self, plan: QueryPlan | None) -> SourceTraits | None:
        """Return cached or inferred source traits for ``plan`` when available."""

        effective_plan = plan or QueryPlan()
        key = self._plan_signature(effective_plan)
        with self._lock:
            cached = self._source_traits_cache.get(key)
        if cached is not None:
            return cached

        if self._fetcher is not None:
            return None

        prepared = self._prepare_materializer(effective_plan)
        if prepared is None:
            return None
        _materializer, physical_plan = prepared
        self._record_source_traits(effective_plan, physical_plan)
        with self._lock:
            return self._source_traits_cache.get(key)

    @staticmethod
    def _plan_is_simple(plan: QueryPlan) -> bool:
        if plan.filter_clauses or plan.sort or plan.search_text:
            return False
        return plan.offset == 0 and plan.limit is None

    def _ensure_strategy_for_context(self, context: PlanContext) -> Strategy | None:
        plan = context.plan
        key = self._plan_signature(plan)
        with self._lock:
            cached = self._strategy_cache.get(key)
            if cached is not None:
                self._apply_strategy_locked(cached)
                return cached

        traits = self.get_source_traits(plan)
        if traits is None:
            return None

        strategy = compile_strategy(traits)
        with self._lock:
            cache = self._strategy_cache
            cache[key] = strategy
            cache.move_to_end(key)
            while len(cache) > self.STRATEGY_CACHE_LIMIT:
                cache.popitem(last=False)
            self._apply_strategy_locked(strategy)
        return strategy

    def _apply_strategy_locked(self, strategy: Strategy) -> None:
        self._strategy = strategy
        if not self._streaming_enabled_configured:
            self._streaming_enabled = strategy.mode == "streaming"
        if not self._streaming_batch_rows_configured:
            self._streaming_batch_rows = max(1, strategy.batch_rows)

    def clear(self) -> None:
        """Cancel pending work and drop cached prefetches."""

        with self._lock:
            self._cache.clear()
            self._source_traits_cache.clear()
            self._strategy_cache.clear()
            self._strategy = None
            self._sidecar.clear()
            self._parquet_indexer.clear()

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------
    def _resolve_context(
        self,
        plan: QueryPlan | None,
        columns: Sequence[str],
    ) -> PlanContext | None:
        requested = tuple(columns)
        if self._fetcher is not None:
            sheet_id, generation, plan_hash = self._current_job_metadata(plan)
            effective_plan = plan or QueryPlan()
            base_columns = tuple(self._schema_template().column_names)
            available_set = set(base_columns)
            present = tuple(col for col in requested if col in available_set)
            missing = tuple(col for col in requested if col not in available_set)
            return PlanContext(
                effective_plan,
                present,
                requested,
                missing,
                plan_hash,
                sheet_id,
                generation,
            )

        if self._engine_factory is None or self._columns_getter is None:
            return None

        available_columns = tuple(self._columns_getter())
        row_id_column = self._row_id_column
        if not available_columns:
            sheet_id, generation, plan_hash = self._current_job_metadata(plan)
            effective_plan = plan or QueryPlan()
            missing = tuple(requested)
            return PlanContext(
                effective_plan,
                (),
                requested,
                missing,
                plan_hash,
                sheet_id,
                generation,
            )

        available_set = set(available_columns)
        if row_id_column:
            available_set.add(row_id_column)
        present = tuple(col for col in requested if col in available_set)

        effective_plan = (plan or QueryPlan()).with_limit(None).with_offset(0)

        if effective_plan.projection:
            plan_for_fetch = effective_plan
        else:
            projection: list[str] = []
            for name in present:
                if name not in projection:
                    projection.append(name)
            for column, _ in effective_plan.sort:
                if column in available_set and column not in projection:
                    projection.append(column)
            if not projection:
                projection = list(available_columns)
            plan_for_fetch = effective_plan.with_projection(projection)

        if plan_for_fetch.projection:
            visible_columns = tuple(name for name in present if name in plan_for_fetch.projection)
        else:
            visible_columns = present

        def _dedup(seq: Sequence[str]) -> tuple[str, ...]:
            seen: set[str] = set()
            unique: list[str] = []
            for name in seq:
                if name in seen:
                    continue
                seen.add(name)
                unique.append(name)
            return tuple(unique)

        if row_id_column and row_id_column in available_set:
            projection_for_fetch = (
                list(plan_for_fetch.projection) if plan_for_fetch.projection else []
            )
            if row_id_column not in projection_for_fetch:
                projection_for_fetch.append(row_id_column)
                plan_for_fetch = plan_for_fetch.with_projection(_dedup(projection_for_fetch))

        sheet_id, generation, plan_hash = self._current_job_metadata(plan_for_fetch)
        missing = tuple(col for col in requested if col not in available_set)
        fetch_columns: tuple[str, ...]
        if row_id_column and row_id_column in available_set:
            fetch_columns = _dedup(list(visible_columns) + [row_id_column])
        else:
            fetch_columns = visible_columns
        return PlanContext(
            plan_for_fetch,
            fetch_columns,
            requested,
            missing,
            plan_hash,
            sheet_id,
            generation,
        )

    def _collect_plan_slice(
        self,
        plan: QueryPlan,
        start: int,
        count: int,
        columns: Sequence[str],
    ) -> TableSlice:
        prepared = self._prepare_materializer(plan)
        if prepared is None:
            return self._empty_slice(columns)
        materializer, physical_plan = prepared
        self._record_source_traits(plan, physical_plan)

        try:
            raw = materializer.collect_slice(
                physical_plan,
                start=start,
                length=count,
                columns=tuple(columns),
            )
        except PulkaCoreError:
            raise
        except Exception as exc:  # pragma: no cover - defensive
            msg = "Failed to materialise row slice"
            raise MaterializeError(msg) from exc
        return self._coerce_slice(raw, row_id_column=self._row_id_column)

    def prepare_sidecar_window(
        self,
        context: PlanContext,
        start: int,
        count: int,
        *,
        record_progress: bool,
    ) -> SidecarWindow:
        if self._fetcher is not None:
            return SidecarWindow.identity(start, count)
        with self._lock:
            strategy = self._strategy
        return self._sidecar.prepare_window(
            context,
            start,
            count,
            strategy=strategy,
            record_progress=record_progress,
        )

    def apply_sidecar_window(
        self,
        raw_slice: TableSlice,
        window: SidecarWindow,
        requested_count: int,
    ) -> TableSlice:
        return self._sidecar.apply_window(raw_slice, window, requested_count)

    def _prepare_materializer(self, plan: QueryPlan) -> tuple[MaterializerProtocol, Any] | None:
        adapter = self._engine_adapter()
        materializer = self._materializer
        if adapter is None or materializer is None:
            return None
        try:
            physical_plan = adapter.compile(plan)
        except PulkaCoreError:
            raise
        except Exception as exc:  # pragma: no cover - defensive
            msg = "Failed to compile plan for row slice"
            raise CompileError(msg) from exc
        return materializer, physical_plan

    def _record_source_traits(self, plan: QueryPlan, physical_plan: Any) -> None:
        traits = self._infer_source_traits_from_physical_plan(physical_plan)
        if traits is None:
            return
        key = self._plan_signature(plan)
        with self._lock:
            cache = self._source_traits_cache
            cache[key] = traits
            cache.move_to_end(key)
            while len(cache) > self.SOURCE_TRAITS_CACHE_LIMIT:
                cache.popitem(last=False)
            self._strategy_cache.pop(key, None)

    @staticmethod
    def _infer_source_traits_from_physical_plan(plan: Any) -> SourceTraits | None:
        try:
            from .engine.viewer_engine import infer_source_traits_from_plan
        except Exception:
            return None
        try:
            return infer_source_traits_from_plan(plan)
        except Exception:
            return None

    @staticmethod
    def _plan_signature(plan: QueryPlan) -> str:
        snapshot = plan.snapshot()
        return cast(str, snapshot["hash"])

    def _fetch_slice(
        self,
        context: PlanContext,
        start: int,
        count: int,
        *,
        window: SidecarWindow | None = None,
        record_progress: bool = True,
    ) -> tuple[TableSlice, SidecarWindow]:
        if not context.fetch_columns:
            empty = self._empty_slice(context.fetch_columns)
            return empty, SidecarWindow.identity(start, count)

        if self._fetcher is not None:
            raw = self._fetcher(start, count, context.fetch_columns)
            coerced = self._coerce_slice(raw, row_id_column=self._row_id_column)
            if coerced.start_offset is None:
                coerced = TableSlice(
                    coerced.columns,
                    coerced.schema,
                    start_offset=start,
                    row_ids=coerced.row_ids,
                )
            return coerced, SidecarWindow.identity(start, count)

        effective_window = window or self.prepare_sidecar_window(
            context,
            start,
            count,
            record_progress=record_progress,
        )
        fetch_start = effective_window.fetch_start
        fetch_count = effective_window.fetch_count
        raw = self._collect_plan_slice(
            context.plan,
            fetch_start,
            fetch_count,
            context.fetch_columns,
        )
        if raw.start_offset is None:
            raw = TableSlice(
                raw.columns,
                raw.schema,
                start_offset=fetch_start,
                row_ids=raw.row_ids,
            )
        trimmed = self.apply_sidecar_window(raw, effective_window, count)
        return trimmed, effective_window

    def _engine_adapter(self) -> EngineAdapterProtocol | None:
        if self._engine_factory is None:
            return None
        adapter = self._engine_factory()
        if not isinstance(adapter, EngineAdapterProtocol):
            msg = "engine_factory must return an EngineAdapterProtocol"
            raise TypeError(msg)
        return adapter

    def _empty_slice(self, columns: Sequence[str] | None = None) -> TableSlice:
        template = self._schema_template()
        if columns is None:
            return TableSlice.empty(template.column_names, template.schema)
        return TableSlice.empty(columns, template.schema)

    @staticmethod
    def _coerce_slice(result: Any, *, row_id_column: str | None = None) -> TableSlice:
        if isinstance(result, TableSlice):
            return result
        if result is None:
            return TableSlice.empty()
        if isinstance(result, pl.DataFrame):
            schema = getattr(result, "schema", {})
            return table_slice_from_dataframe(result, schema, row_id_column=row_id_column)
        try:
            frame = pl.DataFrame(result)
        except Exception as exc:  # pragma: no cover - defensive
            msg = (
                "RowProvider requires slice results compatible with TableSlice; "
                f"received {type(result)!r}"
            )
            raise MaterializeError(msg) from exc
        return table_slice_from_dataframe(frame, frame.schema, row_id_column=row_id_column)

    def _schema_template(self) -> TableSlice:
        template = self._empty_template
        if template is None:
            template = self._coerce_slice(
                self._empty_result_factory(), row_id_column=self._row_id_column
            )
            self._empty_template = template
        return template

    def _finalize_slice(
        self,
        context: PlanContext,
        raw_slice: TableSlice,
    ) -> tuple[TableSlice, SliceStatus]:
        status = SliceStatus.OK
        if context.missing_columns:
            status = self._merge_status(status, SliceStatus.PARTIAL)

        template = self._schema_template()
        schema = dict(template.schema)
        schema.update(raw_slice.schema)

        row_id_column = self._row_id_column

        raw_columns = {column.name: column for column in raw_slice.columns}
        raw_height = raw_slice.height

        row_ids = raw_slice.row_ids
        if row_id_column:
            column = raw_columns.pop(row_id_column, None)
            if row_ids is None and column is not None:
                row_ids = getattr(column, "data", None) or column.values

        result_columns: list[TableColumn] = []

        effective_expected = [
            name for name in context.fetch_columns if not row_id_column or name != row_id_column
        ]
        expected_fetch = set(effective_expected)
        missing_from_raw = [name for name in effective_expected if name not in raw_columns]
        if missing_from_raw:
            status = self._merge_status(status, SliceStatus.SCHEMA_MISMATCH)

        extra_columns = [
            name
            for name in raw_columns
            if name not in expected_fetch and (not row_id_column or name != row_id_column)
        ]
        if extra_columns:
            status = self._merge_status(status, SliceStatus.SCHEMA_MISMATCH)

        placeholder_reason = (
            SliceStatus.SCHEMA_MISMATCH
            if status is SliceStatus.SCHEMA_MISMATCH
            else SliceStatus.PARTIAL
        )

        for name in context.requested_columns:
            column = raw_columns.get(name)
            if column is not None:
                result_columns.append(column)
                continue
            dtype = schema.get(name)
            result_columns.append(
                self._placeholder_column(name, raw_height, dtype, placeholder_reason)
            )

        if not result_columns and raw_columns:
            result_columns.extend(raw_columns[name] for name in raw_slice.column_names)

        if row_id_column and row_ids is None and raw_height > 0:
            base_start = raw_slice.start_offset if raw_slice.start_offset is not None else 0
            row_ids = tuple(base_start + idx for idx in range(raw_height))

        final_slice = TableSlice(
            tuple(result_columns),
            schema,
            start_offset=raw_slice.start_offset,
            row_ids=row_ids,
        )
        return final_slice, status

    def _placeholder_column(
        self,
        name: str,
        height: int,
        dtype: Any,
        status: SliceStatus,
    ) -> TableColumn:
        values = tuple(None for _ in range(max(0, height)))
        null_count = len(values)

        label = "⟂ missing" if status is SliceStatus.PARTIAL else "⟂ schema"

        def _display(
            _row: int,
            _abs_row: int,
            _value: Any,
            _width: int | None,
            *,
            _label: str = label,
        ) -> str:
            return _label

        return TableColumn(name, values, dtype, null_count, _display)

    @staticmethod
    def _merge_status(left: SliceStatus, right: SliceStatus) -> SliceStatus:
        if right is SliceStatus.SCHEMA_MISMATCH:
            return SliceStatus.SCHEMA_MISMATCH
        if right is SliceStatus.PARTIAL and left is SliceStatus.OK:
            return SliceStatus.PARTIAL
        return left

    @staticmethod
    def _cell_count(slice_: TableSlice) -> int:
        return slice_.height * len(slice_.column_names)

    @staticmethod
    def _augment_telemetry(
        payload: dict[str, Any],
        *,
        strategy_payload: dict[str, Any] | None,
        window: SidecarWindow | None = None,
    ) -> dict[str, Any]:
        if strategy_payload is not None:
            payload["strategy"] = strategy_payload
        if window is not None and window.used:
            payload["sidecar"] = "checkpoints_used"
            if window.checkpoint_row is not None:
                payload["sidecar_checkpoint_row"] = window.checkpoint_row
        return payload

    def update_streaming_metrics(
        self,
        *,
        mode: str,
        chunks: int,
        rows: int,
        cells: int,
        duration_ns: int,
    ) -> None:
        with self._lock:
            self._streaming_last_mode = mode
            self._streaming_last_chunks = max(0, chunks)
            self._streaming_last_rows = max(0, rows)
            self._streaming_last_cells = max(0, cells)
            self._streaming_last_duration_ns = max(0, duration_ns)

    def cache_metrics(self) -> dict[str, int | str]:
        """Return current cache occupancy and eviction counters."""

        metrics: dict[str, int | str] = dict(self._cache.cache_metrics())
        with self._lock:
            metrics.update(
                {
                    "streaming_last_mode": self._streaming_last_mode,
                    "streaming_last_chunks": self._streaming_last_chunks,
                    "streaming_last_rows": self._streaming_last_rows,
                    "streaming_last_cells": self._streaming_last_cells,
                    "streaming_last_duration_ns": self._streaming_last_duration_ns,
                }
            )
        return metrics

    @property
    def _page_cache(self) -> OrderedDict[RowKey, TableSlice]:
        return self._cache.page_cache

    def _current_job_metadata(
        self, plan: QueryPlan | None
    ) -> tuple[str | None, int | None, str | None]:
        if self._job_context is None:
            plan_hash = plan.snapshot()["hash"] if plan is not None else None
            return None, None, plan_hash

        sheet_id, generation, plan_hash = self._job_context()
        if plan_hash is None and plan is not None:
            plan_hash = plan.snapshot()["hash"]
        return sheet_id, generation, plan_hash

    def _schedule_prefetch_windows(
        self,
        *,
        plan: QueryPlan | None,
        columns: Sequence[str],
        start: int,
        count: int,
        windows: int,
    ) -> None:
        if windows <= 0 or count <= 0:
            return
        for window_index in range(1, windows + 1):
            next_start = start + window_index * count
            self.prefetch(plan, columns, next_start, count)

    @staticmethod
    def _cache_key(
        plan_hash: str | None,
        start: int,
        count: int,
        columns: Sequence[str],
    ) -> RowKey:
        cols_sig = normalized_columns_key(columns)
        return plan_hash, start, count, cols_sig


__all__ = [
    "CellPreview",
    "RowProvider",
    "SliceStatus",
    "SliceStreamRequest",
    "TableSliceChunk",
]



"""Streaming pipeline for row provider slices."""

from __future__ import annotations

import inspect
import logging
from collections.abc import Iterator, Sequence
from dataclasses import asdict
from time import monotonic_ns
from typing import TYPE_CHECKING, Any, Protocol

import polars as pl

from .engine.contracts import TableSlice
from .engine.polars_adapter import table_slice_from_dataframe
from .errors import MaterializeError, PulkaCoreError
from .interfaces import MaterializerProtocol
from .row_provider_sidecar import SidecarWindow
from .row_provider_types import PlanContext, SliceStatus, SliceStreamRequest, TableSliceChunk
from .strategy import Strategy

if TYPE_CHECKING:
    from .row_provider_cache import RowCacheManager
    from .row_provider_parquet import ParquetRowGroupIndexer
    from .row_provider_sidecar import SidecarWindowManager
    from .source_traits import SourceTraits


LOGGER = logging.getLogger(__name__)

_PARQUET_ROW_GROUP_PARAM: str | None = None
try:  # pragma: no cover - depends on polars version
    _params = inspect.signature(pl.read_parquet).parameters
    if "row_groups" in _params:
        _PARQUET_ROW_GROUP_PARAM = "row_groups"
    elif "row_group" in _params:
        _PARQUET_ROW_GROUP_PARAM = "row_group"
except Exception:
    _PARQUET_ROW_GROUP_PARAM = None


class RowStreamingHost(Protocol):
    _fetcher: Any
    _row_id_column: str | None
    _streaming_enabled: bool
    _streaming_batch_rows: int
    _page_size: int
    _cache: RowCacheManager
    _parquet_indexer: ParquetRowGroupIndexer
    _sidecar: SidecarWindowManager

    def _resolve_context(self, plan: Any, columns: Sequence[str]) -> PlanContext | None: ...

    def _empty_slice(self, columns: Sequence[str] | None = None) -> TableSlice: ...

    def _finalize_slice(
        self, context: PlanContext, raw_slice: TableSlice
    ) -> tuple[TableSlice, SliceStatus]: ...

    def _cache_key(
        self, plan_hash: str | None, start: int, count: int, columns: Sequence[str]
    ) -> Any: ...

    def _fetch_slice(
        self,
        context: PlanContext,
        start: int,
        count: int,
        *,
        window: SidecarWindow | None = None,
        record_progress: bool = True,
    ) -> tuple[TableSlice, SidecarWindow]: ...

    def _prepare_materializer(self, plan: Any) -> tuple[MaterializerProtocol, Any] | None: ...

    def _record_source_traits(self, plan: Any, physical_plan: Any) -> None: ...

    def _coerce_slice(self, result: Any, *, row_id_column: str | None = None) -> TableSlice: ...

    def get_source_traits(self, plan: Any) -> SourceTraits | None: ...

    def _plan_is_simple(self, plan: Any) -> bool: ...

    def _ensure_strategy_for_context(self, context: PlanContext) -> Strategy | None: ...

    def current_strategy(self) -> Strategy | None: ...

    def prepare_sidecar_window(
        self, context: PlanContext, start: int, count: int, *, record_progress: bool
    ) -> SidecarWindow: ...

    def apply_sidecar_window(
        self, raw_slice: TableSlice, window: SidecarWindow, requested_count: int
    ) -> TableSlice: ...

    def _augment_telemetry(
        self,
        payload: dict[str, Any],
        *,
        strategy_payload: dict[str, Any] | None,
        window: SidecarWindow | None = None,
    ) -> dict[str, Any]: ...

    def update_streaming_metrics(
        self, *, mode: str, chunks: int, rows: int, cells: int, duration_ns: int
    ) -> None: ...

    def _cell_count(self, slice_: TableSlice) -> int: ...

    def _schedule_prefetch_windows(
        self, *, plan: Any, columns: Sequence[str], start: int, count: int, windows: int
    ) -> None: ...


class RowStreamingPipeline:
    def __init__(self, host: RowStreamingHost) -> None:
        self._host = host

    def get_slice_stream(self, request: SliceStreamRequest) -> Iterator[TableSliceChunk]:
        columns = tuple(request.columns)
        request_start = int(request.start)
        count = int(request.count)
        telemetry_base = dict(request.telemetry or {})

        if count <= 0:
            yield from self._stream_empty(request_start, columns, telemetry_base)
            return

        context = self._host._resolve_context(request.plan, columns)
        if context is None:
            yield from self._stream_schema_mismatch(request_start, columns, telemetry_base)
            return

        if not context.fetch_columns:
            yield from self._stream_direct_columns(
                context,
                request,
                telemetry_base,
                columns=columns,
                request_start=request_start,
                count=count,
            )
            return

        parquet_stream = self._stream_parquet_rows(
            context,
            request,
            telemetry_base,
            columns=columns,
            request_start=request_start,
            count=count,
        )
        if parquet_stream is not None:
            yield from parquet_stream
            return

        yield from self._stream_sidecar(
            context,
            request,
            telemetry_base,
            columns=columns,
            request_start=request_start,
            count=count,
        )

    def _stream_empty(
        self,
        request_start: int,
        columns: tuple[str, ...],
        telemetry_base: dict[str, Any],
    ) -> Iterator[TableSliceChunk]:
        empty = self._host._empty_slice(columns)
        status = SliceStatus.PARTIAL if columns else SliceStatus.OK
        telemetry = {
            **telemetry_base,
            "mode": "empty",
            "chunks": 1,
            "rows": 0,
            "cells": 0,
            "duration_ns": 0,
            "offset": request_start,
        }
        self._host.update_streaming_metrics(mode="empty", chunks=1, rows=0, cells=0, duration_ns=0)
        yield TableSliceChunk(request_start, empty, status, True, telemetry)

    def _stream_schema_mismatch(
        self,
        request_start: int,
        columns: tuple[str, ...],
        telemetry_base: dict[str, Any],
    ) -> Iterator[TableSliceChunk]:
        empty = self._host._empty_slice(columns)
        telemetry = {
            **telemetry_base,
            "mode": "schema_mismatch",
            "chunks": 1,
            "rows": 0,
            "cells": 0,
            "duration_ns": 0,
            "offset": request_start,
        }
        self._host.update_streaming_metrics(
            mode="schema_mismatch",
            chunks=1,
            rows=0,
            cells=0,
            duration_ns=0,
        )
        yield TableSliceChunk(request_start, empty, SliceStatus.SCHEMA_MISMATCH, True, telemetry)

    def _stream_direct_columns(
        self,
        context: PlanContext,
        request: SliceStreamRequest,
        telemetry_base: dict[str, Any],
        *,
        columns: tuple[str, ...],
        request_start: int,
        count: int,
    ) -> Iterator[TableSliceChunk]:
        key = self._host._cache_key(context.plan_hash, request_start, count, context.fetch_columns)
        cached = self._host._cache.get_cache(key)
        strategy = self._host.current_strategy()
        strategy_payload = asdict(strategy) if strategy is not None else None

        if cached is not None:
            final_slice, status = self._host._finalize_slice(context, cached)
            cells = self._host._cell_count(final_slice)
            telemetry = self._host._augment_telemetry(
                {
                    **telemetry_base,
                    "mode": "cache",
                    "chunks": 1,
                    "rows": final_slice.height,
                    "cells": cells,
                    "duration_ns": 0,
                    "offset": request_start,
                    "plan_hash": context.plan_hash,
                },
                strategy_payload=strategy_payload,
            )
            self._host.update_streaming_metrics(
                mode="cache",
                chunks=1,
                rows=final_slice.height,
                cells=cells,
                duration_ns=0,
            )
            self._host._schedule_prefetch_windows(
                plan=request.plan,
                columns=columns,
                start=request_start,
                count=count,
                windows=strategy.prefetch_windows if strategy is not None else 0,
            )
            yield TableSliceChunk(request_start, final_slice, status, True, telemetry)
            return

        raw = self._host._empty_slice(context.fetch_columns)
        final_slice, status = self._host._finalize_slice(context, raw)
        cells = self._host._cell_count(final_slice)
        telemetry = self._host._augment_telemetry(
            {
                **telemetry_base,
                "mode": "empty",
                "chunks": 1,
                "rows": final_slice.height,
                "cells": cells,
                "duration_ns": 0,
                "offset": request_start,
                "plan_hash": context.plan_hash,
            },
            strategy_payload=strategy_payload,
        )
        self._host._cache.store_cache(key, raw)
        self._host.update_streaming_metrics(
            mode="empty",
            chunks=1,
            rows=final_slice.height,
            cells=cells,
            duration_ns=0,
        )
        yield TableSliceChunk(request_start, final_slice, status, True, telemetry)

    def _stream_parquet_rows(
        self,
        context: PlanContext,
        request: SliceStreamRequest,
        telemetry_base: dict[str, Any],
        *,
        columns: tuple[str, ...],
        request_start: int,
        count: int,
    ) -> Iterator[TableSliceChunk] | None:
        if self._host._fetcher is not None or _PARQUET_ROW_GROUP_PARAM is None:
            return None
        if request.streaming_enabled is False:
            return None
        if not self._host._plan_is_simple(context.plan):
            return None

        traits = self._host.get_source_traits(context.plan)
        if traits is None or traits.kind != "parquet" or not traits.path:
            return None

        index = self._host._parquet_indexer.get(traits.path)
        if index is None or not index.row_group_rows:
            return None

        if count <= 0:
            return None

        def _iterator() -> Iterator[TableSliceChunk]:
            request_start_local = max(0, int(request_start))
            if request_start_local >= index.total_rows:
                empty = self._host._empty_slice(context.fetch_columns)
                status = SliceStatus.PARTIAL if context.requested_columns else SliceStatus.OK
                telemetry = self._host._augment_telemetry(
                    {
                        **telemetry_base,
                        "mode": "stream",
                        "reason": "row_group_seek",
                        "chunks": 1,
                        "rows": 0,
                        "cells": 0,
                        "duration_ns": 0,
                        "offset": request_start_local,
                        "plan_hash": context.plan_hash,
                    },
                    strategy_payload=None,
                )
                self._host.update_streaming_metrics(
                    mode="stream",
                    chunks=1,
                    rows=0,
                    cells=0,
                    duration_ns=0,
                )
                yield TableSliceChunk(request_start_local, empty, status, True, telemetry)
                return

            strategy = self._host._ensure_strategy_for_context(context)
            strategy_payload = asdict(strategy) if strategy is not None else None
            prefetch_windows = strategy.prefetch_windows if strategy is not None else 0
            prefetch_scheduled = False

            batch_rows = (
                request.batch_rows
                if request.batch_rows and request.batch_rows > 0
                else self._host._streaming_batch_rows
            )

            parquet_path = traits.path
            if parquet_path is None:  # pragma: no cover - defensive
                raise RuntimeError("Parquet streaming requested without a dataset path")

            def _read_group(group_index: int) -> pl.DataFrame:
                kwargs: dict[str, Any] = {}
                if context.fetch_columns:
                    source_columns = [
                        name for name in context.fetch_columns if name != self._host._row_id_column
                    ]
                    if source_columns:
                        kwargs["columns"] = source_columns
                if _PARQUET_ROW_GROUP_PARAM == "row_groups":
                    kwargs["row_groups"] = [group_index]
                elif _PARQUET_ROW_GROUP_PARAM == "row_group":
                    kwargs["row_group"] = group_index
                return pl.read_parquet(parquet_path, **kwargs)

            start_ns = monotonic_ns()
            last_chunk_ns = start_ns
            total_rows = 0
            chunk_index = 0
            remaining = max(0, int(count))
            start_group, group_start = index.seek(request_start_local)
            trim_leading = max(0, request_start_local - group_start)
            assembled_raw: TableSlice | None = None

            try:
                for group_idx in range(start_group, len(index.row_group_rows)):
                    if remaining <= 0:
                        break
                    group_start = index.row_group_offsets[group_idx]
                    df = _read_group(group_idx)
                    if trim_leading:
                        df = df.slice(trim_leading, None)
                        group_start += trim_leading
                        trim_leading = 0
                    if remaining < df.height:
                        df = df.slice(0, remaining)
                    if df.height <= 0:
                        continue

                    if self._host._row_id_column:
                        row_ids = pl.Series(
                            self._host._row_id_column,
                            range(group_start, group_start + df.height),
                        )
                        df = df.with_columns(row_ids)

                    raw_slice = table_slice_from_dataframe(
                        df,
                        df.schema,
                        row_id_column=self._host._row_id_column,
                    )
                    if raw_slice.start_offset is None:
                        raw_slice = TableSlice(
                            raw_slice.columns,
                            raw_slice.schema,
                            start_offset=group_start,
                            row_ids=raw_slice.row_ids,
                        )

                    chunk_slices = [raw_slice]
                    if batch_rows and raw_slice.height > batch_rows:
                        chunk_slices = []
                        offset = 0
                        while offset < raw_slice.height:
                            length = min(batch_rows, raw_slice.height - offset)
                            chunk_slices.append(raw_slice.slice(offset, length))
                            offset += length

                    for chunk_slice in chunk_slices:
                        if chunk_slice.height <= 0:
                            continue
                        chunk_offset = chunk_slice.start_offset or (
                            request_start_local + total_rows
                        )
                        total_rows += chunk_slice.height
                        remaining = max(0, count - total_rows)
                        assembled_raw = (
                            chunk_slice
                            if assembled_raw is None
                            else assembled_raw.concat_vertical(chunk_slice)
                        )
                        finalized_chunk, chunk_status = self._host._finalize_slice(
                            context, chunk_slice
                        )
                        now_ns = monotonic_ns()
                        chunk_duration_ns = now_ns - last_chunk_ns
                        last_chunk_ns = now_ns
                        chunk_index += 1
                        chunk_cells = self._host._cell_count(finalized_chunk)
                        chunk_telemetry = self._host._augment_telemetry(
                            {
                                **telemetry_base,
                                "mode": "stream",
                                "reason": "row_group_seek",
                                "chunk_index": chunk_index,
                                "rows": finalized_chunk.height,
                                "cells": chunk_cells,
                                "duration_ns": chunk_duration_ns,
                                "offset": chunk_offset,
                                "plan_hash": context.plan_hash,
                                "row_group_start": group_start,
                                "row_group_index": group_idx,
                            },
                            strategy_payload=strategy_payload,
                        )
                        yield TableSliceChunk(
                            chunk_offset,
                            finalized_chunk,
                            chunk_status,
                            False,
                            chunk_telemetry,
                        )
                        if not prefetch_scheduled and prefetch_windows > 0:
                            self._host._schedule_prefetch_windows(
                                plan=request.plan,
                                columns=columns,
                                start=request_start_local,
                                count=count,
                                windows=prefetch_windows,
                            )
                            prefetch_scheduled = True
                        if remaining <= 0:
                            break
            except Exception as exc:
                msg = "Failed to stream parquet row groups"
                raise MaterializeError(msg) from exc

            if assembled_raw is None:
                assembled_raw = self._host._empty_slice(context.fetch_columns)

            total_duration_ns = monotonic_ns() - start_ns
            final_slice, status = self._host._finalize_slice(context, assembled_raw)
            final_cells = self._host._cell_count(final_slice)
            total_chunks = max(1, chunk_index + 1)
            summary_telemetry = self._host._augment_telemetry(
                {
                    **telemetry_base,
                    "mode": "stream",
                    "reason": "row_group_seek",
                    "chunk_index": total_chunks,
                    "chunks": total_chunks,
                    "rows": final_slice.height,
                    "cells": final_cells,
                    "duration_ns": total_duration_ns,
                    "offset": request_start_local,
                    "plan_hash": context.plan_hash,
                    "row_group_index": start_group,
                },
                strategy_payload=strategy_payload,
            )
            key = self._host._cache_key(
                context.plan_hash,
                request_start_local,
                count,
                context.fetch_columns,
            )
            self._host._cache.store_cache(key, assembled_raw)
            self._host.update_streaming_metrics(
                mode="stream",
                chunks=total_chunks,
                rows=final_slice.height,
                cells=final_cells,
                duration_ns=total_duration_ns,
            )
            if not prefetch_scheduled and prefetch_windows > 0:
                self._host._schedule_prefetch_windows(
                    plan=request.plan,
                    columns=columns,
                    start=request_start_local,
                    count=count,
                    windows=prefetch_windows,
                )
            yield TableSliceChunk(request_start_local, final_slice, status, True, summary_telemetry)

        return _iterator()

    def _stream_sidecar(
        self,
        context: PlanContext,
        request: SliceStreamRequest,
        telemetry_base: dict[str, Any],
        *,
        columns: tuple[str, ...],
        request_start: int,
        count: int,
    ) -> Iterator[TableSliceChunk]:
        strategy = self._host._ensure_strategy_for_context(context)
        window = SidecarWindow.identity(request_start, count)
        if self._host._fetcher is None:
            window = self._host.prepare_sidecar_window(
                context,
                request_start,
                count,
                record_progress=True,
            )

        strategy_payload = asdict(strategy) if strategy is not None else None

        key = self._host._cache_key(context.plan_hash, request_start, count, context.fetch_columns)
        cached = self._host._cache.get_cache(key)
        if cached is not None:
            final_slice, status = self._host._finalize_slice(context, cached)
            cells = self._host._cell_count(final_slice)
            telemetry = self._host._augment_telemetry(
                {
                    **telemetry_base,
                    "mode": "cache",
                    "chunks": 1,
                    "rows": final_slice.height,
                    "cells": cells,
                    "duration_ns": 0,
                    "offset": request_start,
                    "plan_hash": context.plan_hash,
                },
                strategy_payload=strategy_payload,
                window=window,
            )
            self._host.update_streaming_metrics(
                mode="cache",
                chunks=1,
                rows=final_slice.height,
                cells=cells,
                duration_ns=0,
            )
            self._host._schedule_prefetch_windows(
                plan=request.plan,
                columns=columns,
                start=request_start,
                count=count,
                windows=strategy.prefetch_windows if strategy is not None else 0,
            )
            yield TableSliceChunk(request_start, final_slice, status, True, telemetry)
            return

        stream_enabled = (
            request.streaming_enabled
            if request.streaming_enabled is not None
            else self._host._streaming_enabled
        )
        batch_rows = (
            request.batch_rows
            if request.batch_rows and request.batch_rows > 0
            else self._host._streaming_batch_rows
        )

        if self._host._fetcher is not None:
            start_ns = monotonic_ns()
            raw_slice, _ = self._host._fetch_slice(
                context,
                request_start,
                count,
                record_progress=False,
            )
            final_slice, status = self._host._finalize_slice(context, raw_slice)
            duration_ns = monotonic_ns() - start_ns
            cells = self._host._cell_count(final_slice)
            telemetry = self._host._augment_telemetry(
                {
                    **telemetry_base,
                    "mode": "passthrough",
                    "chunks": 1,
                    "rows": final_slice.height,
                    "cells": cells,
                    "duration_ns": duration_ns,
                    "offset": request_start,
                    "plan_hash": context.plan_hash,
                },
                strategy_payload=strategy_payload,
            )
            self._host._cache.store_cache(key, raw_slice)
            self._host.update_streaming_metrics(
                mode="passthrough",
                chunks=1,
                rows=final_slice.height,
                cells=cells,
                duration_ns=duration_ns,
            )
            self._host._schedule_prefetch_windows(
                plan=request.plan,
                columns=columns,
                start=request_start,
                count=count,
                windows=strategy.prefetch_windows if strategy is not None else 0,
            )
            yield TableSliceChunk(request_start, final_slice, status, True, telemetry)
            return

        start_ns = monotonic_ns()
        prepared = self._host._prepare_materializer(context.plan)
        if prepared is None:
            raw_slice = self._host._empty_slice(context.fetch_columns)
            final_slice, status = self._host._finalize_slice(context, raw_slice)
            cells = self._host._cell_count(final_slice)
            telemetry = self._host._augment_telemetry(
                {
                    **telemetry_base,
                    "mode": "collect",
                    "chunks": 1,
                    "rows": final_slice.height,
                    "cells": cells,
                    "duration_ns": 0,
                    "offset": request_start,
                    "plan_hash": context.plan_hash,
                },
                strategy_payload=strategy_payload,
                window=window,
            )
            self._host._cache.store_cache(key, raw_slice)
            self._host.update_streaming_metrics(
                mode="collect",
                chunks=1,
                rows=final_slice.height,
                cells=cells,
                duration_ns=0,
            )
            self._host._schedule_prefetch_windows(
                plan=request.plan,
                columns=columns,
                start=request_start,
                count=count,
                windows=strategy.prefetch_windows if strategy is not None else 0,
            )
            yield TableSliceChunk(request_start, final_slice, status, True, telemetry)
            return

        materializer, physical_plan = prepared
        self._host._record_source_traits(context.plan, physical_plan)
        stream_attr = getattr(materializer, "collect_slice_stream", None)
        stream_iterator: Iterator[TableSlice] | None = None
        stream_mode = "collect"
        stream_reason = "disabled"
        prefetch_windows = strategy.prefetch_windows if strategy is not None else 0
        prefetch_scheduled = False
        stream_kwargs = {
            "start": window.fetch_start,
            "length": window.fetch_count,
            "columns": tuple(context.fetch_columns),
        }
        if stream_enabled and callable(stream_attr):
            if batch_rows > 0:
                stream_kwargs["batch_rows"] = batch_rows
            try:
                stream_iterator = iter(stream_attr(physical_plan, **stream_kwargs))
                stream_mode = "stream"
                stream_reason = "stream"
            except TypeError:
                stream_kwargs.pop("batch_rows", None)
                try:
                    stream_iterator = iter(stream_attr(physical_plan, **stream_kwargs))
                    stream_mode = "stream"
                    stream_reason = "stream"
                except TypeError:
                    stream_iterator = None
                    stream_reason = "type_error"
                except PulkaCoreError:
                    raise
                except Exception:
                    stream_iterator = None
                    stream_reason = "error"
            except PulkaCoreError:
                raise
            except Exception:
                stream_iterator = None
                stream_reason = "error"
        else:
            if not stream_enabled:
                stream_reason = "disabled"
            elif stream_attr is None:
                stream_reason = "missing"
            else:
                stream_reason = "invalid"

        if stream_iterator is None:
            try:
                raw = materializer.collect_slice(
                    physical_plan,
                    start=window.fetch_start,
                    length=window.fetch_count,
                    columns=tuple(context.fetch_columns),
                )
            except PulkaCoreError:
                raise
            except Exception as exc:  # pragma: no cover - defensive
                msg = "Failed to materialise row slice"
                raise MaterializeError(msg) from exc
            coerced = self._host._coerce_slice(raw, row_id_column=self._host._row_id_column)
            trimmed = self._host.apply_sidecar_window(coerced, window, count)
            final_slice, status = self._host._finalize_slice(context, trimmed)
            duration_ns = monotonic_ns() - start_ns
            cells = self._host._cell_count(final_slice)
            telemetry = self._host._augment_telemetry(
                {
                    **telemetry_base,
                    "mode": stream_mode,
                    "reason": stream_reason,
                    "chunks": 1,
                    "rows": final_slice.height,
                    "cells": cells,
                    "duration_ns": duration_ns,
                    "offset": request_start,
                    "plan_hash": context.plan_hash,
                },
                strategy_payload=strategy_payload,
                window=window,
            )
            self._host._cache.store_cache(key, trimmed)
            self._host.update_streaming_metrics(
                mode=stream_mode,
                chunks=1,
                rows=final_slice.height,
                cells=cells,
                duration_ns=duration_ns,
            )
            yield TableSliceChunk(request_start, final_slice, status, True, telemetry)
            return

        total_rows = 0
        chunk_index = 0
        last_chunk_ns = start_ns
        skip_rows = window.trim_leading
        assembled_raw: TableSlice | None = None
        try:
            while True:
                try:
                    raw_chunk = next(stream_iterator)
                except StopIteration:
                    break
                chunk_slice = self._host._coerce_slice(
                    raw_chunk, row_id_column=self._host._row_id_column
                )
                if chunk_slice.height <= 0:
                    continue
                if skip_rows > 0:
                    if chunk_slice.height <= skip_rows:
                        skip_rows -= chunk_slice.height
                        continue
                    chunk_slice = chunk_slice.slice(skip_rows, None)
                    skip_rows = 0
                remaining = max(0, count - total_rows)
                if remaining <= 0:
                    break
                if chunk_slice.height > remaining:
                    chunk_slice = chunk_slice.slice(0, remaining)
                if chunk_slice.height <= 0:
                    continue
                chunk_offset = request_start + total_rows
                total_rows += chunk_slice.height
                if chunk_slice.start_offset is None:
                    chunk_slice = TableSlice(
                        tuple(chunk_slice.columns),
                        chunk_slice.schema,
                        start_offset=chunk_offset,
                        row_ids=chunk_slice.row_ids,
                    )
                assembled_raw = (
                    chunk_slice
                    if assembled_raw is None
                    else assembled_raw.concat_vertical(chunk_slice)
                )
                finalized_chunk, chunk_status = self._host._finalize_slice(context, chunk_slice)
                now_ns = monotonic_ns()
                chunk_duration_ns = now_ns - last_chunk_ns
                last_chunk_ns = now_ns
                chunk_index += 1
                chunk_cells = self._host._cell_count(finalized_chunk)
                chunk_telemetry = self._host._augment_telemetry(
                    {
                        **telemetry_base,
                        "mode": stream_mode,
                        "reason": stream_reason,
                        "chunk_index": chunk_index,
                        "rows": finalized_chunk.height,
                        "cells": chunk_cells,
                        "duration_ns": chunk_duration_ns,
                        "offset": chunk_offset,
                        "plan_hash": context.plan_hash,
                    },
                    strategy_payload=strategy_payload,
                    window=window,
                )
                if LOGGER.isEnabledFor(logging.DEBUG):
                    LOGGER.debug(
                        "row_provider.stream_chunk",
                        extra={
                            "event": "row_stream_chunk",
                            "plan_hash": context.plan_hash,
                            "chunk_index": chunk_index,
                            "rows": finalized_chunk.height,
                            "offset": chunk_offset,
                        },
                    )
                yield TableSliceChunk(
                    chunk_offset,
                    finalized_chunk,
                    chunk_status,
                    False,
                    chunk_telemetry,
                )
                if not prefetch_scheduled and prefetch_windows > 0:
                    self._host._schedule_prefetch_windows(
                        plan=request.plan,
                        columns=columns,
                        start=request_start,
                        count=count,
                        windows=prefetch_windows,
                    )
                    prefetch_scheduled = True
                chunk_offset += chunk_slice.height
                if total_rows >= count:
                    break
        except PulkaCoreError:
            raise
        except Exception as exc:  # pragma: no cover - defensive
            msg = "Failed to materialise streaming slice"
            raise MaterializeError(msg) from exc

        if assembled_raw is None:
            assembled_raw = self._host._empty_slice(context.fetch_columns)
        total_duration_ns = monotonic_ns() - start_ns
        final_slice, status = self._host._finalize_slice(context, assembled_raw)
        final_cells = self._host._cell_count(final_slice)
        total_chunks = chunk_index + 1
        summary_telemetry = self._host._augment_telemetry(
            {
                **telemetry_base,
                "mode": stream_mode,
                "reason": stream_reason,
                "chunk_index": total_chunks,
                "chunks": total_chunks,
                "rows": final_slice.height,
                "cells": final_cells,
                "duration_ns": total_duration_ns,
                "offset": request_start,
                "plan_hash": context.plan_hash,
            },
            strategy_payload=strategy_payload,
            window=window,
        )
        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug(
                "row_provider.stream_complete",
                extra={
                    "event": "row_stream_complete",
                    "plan_hash": context.plan_hash,
                    "chunks": total_chunks,
                    "rows": final_slice.height,
                    "duration_ns": total_duration_ns,
                },
            )
        self._host._cache.store_cache(key, assembled_raw)
        self._host.update_streaming_metrics(
            mode=stream_mode,
            chunks=total_chunks,
            rows=final_slice.height,
            cells=final_cells,
            duration_ns=total_duration_ns,
        )
        if not prefetch_scheduled and prefetch_windows > 0:
            self._host._schedule_prefetch_windows(
                plan=request.plan,
                columns=columns,
                start=request_start,
                count=count,
                windows=prefetch_windows,
            )
        yield TableSliceChunk(request_start, final_slice, status, True, summary_telemetry)


__all__ = ["RowStreamingPipeline"]
